@@ -1,5 +1,5 @@
 (*
-   Copyright 2007-2011 by Mark Weyer
+   Copyright 2007-2011,2014 by Mark Weyer
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -76,6 +76,20 @@ let xpm_of_rgba ?(quant_colours=None) ?(quant_method="maximal") file = [
     " -"^quant_method^
     " "^file]]
 
+let xpm_of_rgbas ?(quant_colours=None) ?(quant_method="maximal")
+    width files file =
+  [
+    [file^".xpm"],
+    "machxpm" :: List.map (fun f -> f^".pam") files,
+    ["./machxpm -rgba "^
+      (match quant_colours  with
+      | None -> ""
+      | Some n -> "-colours "^(string_of_int n))^
+      " -"^quant_method^
+      " -width "^(string_of_int width)^
+      " -o "^file^".xpm "^
+      String.concat " " files]]
+
 let pam_of_ppmpgm file = [
   [file^".pam"],
   [file^".ppm"; file^".umriss.pgm"],
@@ -98,12 +112,6 @@ let xpm_of_ppm2 quant_colours quant_method file =
     xpm_of_rgba ~quant_colours:quant_colours ~quant_method:quant_method file;
   ]
 
-let pov_umriss file = [
-  [file^".umriss.pov"],
-  [file^".pov"],
-  ["echo \"#declare Nur_Umriss=1;\" | "^
-      "cat - "^file^".pov > "^file^".umriss.pov"]]
-
 let ppm_of_pov width height aa extra file includes =
   let w,h = width*gric, height*gric  in
   let ws,hs = string_of_int w, string_of_int h  in
@@ -113,26 +121,45 @@ let ppm_of_pov width height aa extra file includes =
          Some true for antialising which respects pixel boundaries
          Some false for antialising which does not respect pixel boundaries *)
   [file^".ppm"],
-  [file^".pov"]@includes,
+  [file^".pov";"version.inc"]@includes,
   ["povray +FP -D -w"^ws^" -h"^hs^
       (match aa with
       |	None -> ""
       |	Some true -> " +A +AM1 -J"
       |	Some false -> " +A +AM2 +R2 -J")^
+      " +HIversion.inc"^
       " "^extra^
       " "^file^".pov";
-    "test -e "^file^".ppm && test `wc -c < "^file^".ppm` -eq "^
-        (* Test whether povray did well.
-           (Exit code 0 from povray does not really mean much.) *)
-      (string_of_int (w*h*3+(String.length ws)+(String.length hs)+9));
+    "test -e "^file^".ppm";
+  ]]
+
+let ppm_of_pov_umriss width height aa extra file includes =
+  let w,h = width*gric, height*gric  in
+  let ws,hs = string_of_int w, string_of_int h  in
+  [
+    (* The antialiasing argument aa is:
+         None for no antialiasing
+         Some true for antialising which respects pixel boundaries
+         Some false for antialising which does not respect pixel boundaries *)
+  [file^".umriss.ppm"],
+  [file^".pov";"umriss.inc"]@includes,
+  ["povray +FP -D -w"^ws^" -h"^hs^
+      (match aa with
+      |	None -> ""
+      |	Some true -> " +A +AM1 -J"
+      |	Some false -> " +A +AM2 +R2 -J")^
+      " +HIumriss.inc"^
+      " -O"^file^".umriss.ppm"^
+      " "^extra^
+      " "^file^".pov";
+    "test -e "^file^".umriss.ppm";
   ]]
 
 let pam_of_pov width height aa extra file includes =
   let includes = "cuyopov.inc"::includes  in
   List.concat [
     ppm_of_pov width height aa extra file includes;
-    pov_umriss file;
-    ppm_of_pov width height aa extra (file^".umriss") includes;
+    ppm_of_pov_umriss width height aa extra file includes;
     pam_of_ppm2 file;
   ]
 
@@ -247,17 +274,58 @@ let xpm_of_pov_fill width height ?(aa=Some false) ?(extra="")
     pov_fill file povvar stage])
   stages)
 
+let gimp_script_xpm_of_xcf threshold used_layers srcname dstname =
+  "(let* (" ^
+    "(image (car (gimp-xcf-load 0 \"" ^ srcname ^ "\" \"" ^ srcname ^ "\")))" ^
+    "(layers_with_num (gimp-image-get-layers image))" ^
+    "(num_layers (car layers_with_num))" ^
+    "(layers (cadr layers_with_num))" ^
+    "(current_layer 0)" ^
+    "(used_layers (list " ^
+      String.concat " " (List.map string_of_int used_layers) ^ " 10000000))" ^
+    "(visible TRUE)" ^
+    ")" ^
+  "(while (< current_layer num_layers)" ^
+    "(if (= current_layer (car used_layers))" ^
+      "(set! visible TRUE)" ^
+      "(set! visible FALSE))" ^
+    "(gimp-item-set-visible (aref layers current_layer) visible)" ^
+    "(if (= visible TRUE) (set! used_layers (cdr used_layers)))" ^
+    "(set! current_layer (+ current_layer 1))" ^
+    ")" ^
+  "(file-xpm-save RUN-NONINTERACTIVE image" ^
+    "(car (gimp-image-merge-visible-layers image CLIP-TO-IMAGE))" ^
+    "\"" ^ dstname ^ "\" \"" ^ dstname ^ "\" " ^ string_of_int threshold^ "))" ^
+  "(gimp-quit TRUE)"
+
+let xpm_of_xcf ?(threshold=127) used_layers ?srcname dstname =
+  let srcname = match srcname with
+  | None -> dstname
+  | Some s -> s  in
+  let srcname = srcname ^ ".xcf"  in
+  let dstname = dstname ^ ".xpm"  in
+  [
+    [dstname],
+    [srcname],
+    ["gimp -i -d -f -b '" ^
+      gimp_script_xpm_of_xcf threshold used_layers srcname dstname ^ "'"]]
+
+let xpms_of_xcf ?threshold srcname specs = List.concat (List.map
+  (fun (dstname,used_layers) ->
+    xpm_of_xcf ?threshold used_layers ~srcname:srcname dstname)
+  specs)
+
 (*===========================================================================*)
 
 let rules = List.concat [
 
     group "all"
       ["aehnlich"; "angst"; "augen"; "aux"; "bonimali";
-        "bunt"; "dungeon"; "fische"; "flechtwerk"; "jump";
+        "bunt"; "dungeon"; "fische"; "flechtwerk"; "fractals"; "jump";
         "kacheln"; "kolben"; "novips";
         "octopi"; "puzzle"; "rechnen"; "reversi_brl";
         "rollenspiel"; "schemen"; "secret"; "slime";
-        "tennis"; "trees"; "zahn"; "ziehlen"];
+        "tennis"; "trees"; "xcf"; "zahn"; "ziehlen"];
 
 
     xzgroup "aehnlich" ["maeSorten"; "maeSchema"];
@@ -267,7 +335,7 @@ let rules = List.concat [
 
 
     group_of_ml "augen" (fill_anim "ma"
-        ("Leer" :: fill_anim "Lid" ["A";"B";"C";"D"] "" @ num_anim 1 1 5)
+        (fill_anim "Lid" ["A";"B";"C";"D"] "" @ num_anim 1 1 5)
         "")
       ml_vektorgraphik;
 
@@ -326,13 +394,11 @@ let rules = List.concat [
        da sich die Graphiken berlappen, so da antialiasing gegen
        irgendeine feste Farbe immer falsch ist.
     *)
-    (let farben22 =
-      ["Plastik"; "Gold"]  in
-    let farben31 =
+    (let farben31 =
       ["Ziegel"; "Holz"; "Eisen"; "Stein"; "Fels"]  in
     let farben11 =
       ["EgoV"; "EgoH"; "EgoL"; "EgoR"]  in
-    let farben = farben22@farben31@farben11  in
+    let farben = "Plastik"::"Gold"::farben31@farben11  in
     let boden = num_anim 2 0 15  in
     let ziele = num_anim 2 0 11  in
     let render = num_anim 3 0 34  in
@@ -373,34 +439,35 @@ let rules = List.concat [
             teile);
             xpm_of_rgba ("md"^farbe^ziel);
       ])  liste)  in
+    let draufseite width draufheight draufversionen seitenversionen farben =
+      List.concat (List.map
+        (fun farbe ->
+          let zeile versionart v0 v1 h = List.concat (List.map
+            (fun version -> List.concat [
+              pam_of_pov 1 h None ""
+                ("md"^farbe^version) (includes farbe);
+              pov_fill ("md"^farbe) versionart version;
+              ])
+            (num_anim 1 v0 v1))  in
+          List.concat [
+            xpm_of_rgbas (width*gric)
+              (List.map (fun v -> "md"^farbe^v)
+                (num_anim 1 0 (draufversionen+seitenversionen-1)))
+              ("md"^farbe);
+            zeile "DraufVersion" 0 (draufversionen-1) draufheight;
+            zeile "SeitenVersion" draufversionen
+              (draufversionen+seitenversionen-1) 1;
+            ])
+        farben)  in
     List.concat [
       group "dungeon" (fill_anim "dungeon" ("Boden"::"Himmel"::farben) "");
       group "dungeonEgo" (fill_anim "dungeonEgo" ["V";"H";"L";"R"] "");
       xzgroup "dungeonBoden" (fill_anim "mdBoden" (""::boden) "");
       xzgroup "dungeonHimmel" (fill_anim "mdHimmel" himmel "");
-      List.concat (List.map
-        (fun farbe -> 
-          let drehungen = num_anim 1 0 2  in
-          List.concat [
-            xpm_of_rgba ("md"^farbe);
-            [gruppe farbe "" drehungen];
-            List.concat (List.map
-              (fun drehung -> List.concat [
-                pam_of_pov 3 1 (Some true) ""
-                  ("md"^farbe^drehung) (includes farbe);
-                pov_fill ("md"^farbe) "DraufsichtDrehung" drehung;
-                ])
-              drehungen);
-            ])
-        farben31);
-      List.concat (List.map
-        (fun farbe ->
-          xpm_of_pov_trans 2 2 ~aa:(Some true) ("md"^farbe) (includes farbe))
-        farben22);
-      List.concat (List.map
-        (fun farbe ->
-          xpm_of_pov_trans 1 1 ~aa:(Some true) ("md"^farbe) (includes farbe))
-        farben11);
+      draufseite 3 2 3 3 farben31;
+      draufseite 2 2 4 0 farben11;
+      draufseite 2 2 4 1 ["Gold"];
+      draufseite 2 1 6 0 ["Plastik"];
       xpm_of_pov 1 1 ~aa:(Some true) "mdBoden" (includes "Boden");
       xpm_of_pov_fill 4 2 "mdHimmel" (includes "Himmel")
         "HimmelVersion" himmel;
@@ -458,6 +525,11 @@ let rules = List.concat [
       ml_vektorgraphik;
 
 
+    group "fractals" ["aDragon.xpm.gz"];
+    [["aDragon.xpm.gz"], ["aDragon.ps"],
+      ["convert aDragon.ps aDragon.xpm.gz"]];
+
+
     xzgroup "jump" (fill_anim "mjZeug" (num_anim 1 1 6) "");
     recolour "mjZeug.src" "mjZeug1" (Some (255,0,0));
     recolour "mjZeug.src" "mjZeug2" (Some (255,255,0));
@@ -556,6 +628,41 @@ let rules = List.concat [
     recolour "mtr.src" "mtr5" (Some (128,128,128));
 
 
+    xzgroup "xcf" [
+      "i3Dreieck"; "i3Gitter"; "i3Grau"; "i3Kreis"; "i3Plus"; "i3Quadrat";
+      "i3Stern";
+      "ipStart";
+      "itrBlau"; "itrBraun"; "itrGrau"; "itrGruen"; "itrLila";
+      "iwaBad"; "iwaBeton"; "iwaNix"; "iwaParkett"; "iwaTeppich";
+      "lrKamin"; "lrWasser"];
+    xpms_of_xcf "i3Bunt" [
+      "i3Dreieck",[0;3];
+      "i3Gitter",[0;1];
+      "i3Kreis",[0;5];
+      "i3Plus",[0;2];
+      "i3Quadrat",[0;6];
+      "i3Stern",[0;4];
+      ];
+    xpm_of_xcf [0;1;2;3] "i3Grau";
+    xpm_of_xcf [0] "ipStart";
+    xpms_of_xcf "itrAlle" [
+      "itrBlau",[0;4];
+      "itrBraun",[2;4];
+      "itrGrau",[4];
+      "itrGruen",[3;4];
+      "itrLila",[1;4];
+      ];
+    xpms_of_xcf "iwaAlles" [
+      "iwaBad",[0;4;5;12];
+      "iwaBeton",[0;6;12];
+      "iwaNix",[0;6;7];
+      "iwaParkett",[0;3;5;12];
+      "iwaTeppich",[0;1;2;12];
+      ];
+    xpm_of_xcf [1;2;3] "lrKamin";
+    xpm_of_xcf [0;1] "lrWasser";
+
+
     xzgroup "zahn" ["mzZahn"; "mzZahnGras"; "mzZahnDreh"];
     xpm_of_pov_trans 3 9 "mzZahn" ["zahn.inc"];
     xpm_of_pov_trans 3 9 "mzZahnGras" ["zahn.inc"];
