@@ -4,6 +4,9 @@
     begin                : Thu Jul 20 2000
     copyright            : (C) 2000 by Immi
     email                : cuyo@pcpool.mathematik.uni-freiburg.de
+
+Modified 2001-2003,2005,2006,2008,2010,2011,2014 by the cuyo developers
+
  ***************************************************************************/
 
 /***************************************************************************
@@ -19,16 +22,12 @@
 
 #include "cuyointl.h"
 #include "blop.h"
-#include "leveldaten.h"
 #include "variable.h"
 #include "code.h"
-#include "blopgitter.h" // nur um zukunftsZuweisung() aufrufen zu knnen...
 #include "knoten.h"
-
 #include "cuyo.h" // Um Punkte senden zu knnen
 #include "spielfeld.h"
-
-
+#include "aufnahme.h"
 
 
 
@@ -37,56 +36,69 @@
 /** Erzeugt einen uninitialisierten Blop (der noch nicht verwendet
     werden kann, aber dafr auch keine Fehler hervorruft, wenn irgend
     was anderes nicht initialisiert ist). Das Uninitialisiert sein
-    eines Blops erkennt man daran, dass mDaten 0 ist. */
+    eines Blops erkennt man daran, dass mDaten 0 ist...
+    *und* mBild uninitialisiert ist. Wenn mBild initialisiert ist,
+    handelt es sich um einen reinen Bildstapel-Blop. */
 Blop::Blop(): mFreilaufend(true), mBesitzer(0),
-  mX(keine_koord), mY(keine_koord),
+  mOrt(absort_nirgends),
   mDaten(0), mDatenAlt(0),
   mZeitNummerDatenAlt(-1),
-  mZeitNummerInitStapel(-1),
-  /* Uninitialisierte Bildstapel erzeugen. Sonst wird versucht, auf
+  mZeitNummerLeereStapel(-1),
+  /* Uninitialisierten Bildstapel erzeugen. Sonst wird versucht, auf
      noch uninitialisierte Felder von ld zuzugreifen. */
-  mBild(false), mBildAlt(false),
-  mWillInitEvent(false)
+  mBild(false), mBildAlt(false), mScheduleEventNr(event_keins)
 {
 }
 
 
 
 /** Constructor... s ist SortenNr. */
-Blop::Blop(int s):
+Blop::Blop(int s, int v /*=-1*/):
   mKettenGroesse(1),
-  mAmPlatzen(0),
-  mFreilaufend(true), mBesitzer(0), mX(keine_koord), mY(keine_koord),
+  mFreilaufend(true), mBesitzer(0), mOrt(absort_nirgends),
   mDatenLaenge(ld->mLevelKnoten->getDatenLaenge()),
   mZeitNummerDatenAlt(-1),
-  mZeitNummerInitStapel(-1),
-  mWillInitEvent(false)
+  mZeitNummerLeereStapel(-1),
+  mScheduleEventNr(event_keins)
 {
   mDaten = new int[mDatenLaenge];
   mDatenAlt = new int[mDatenLaenge];
   /* Zustandsdaten auf Default-Werte initialisieren */
   for (int i = 0; i < mDatenLaenge; i++)
-    mDaten[i] = ld->mLevelKnoten->getDefaultWert(i);
+    switch (ld->mSorten[s]->getDefaultArt(i)) {
+      case da_nie:
+      case da_keinblob:
+	break;
+      case da_init:
+      case da_kind:
+      case da_event:
+	mDaten[i] = ld->mSorten[s]->getDefault(i);
+        break;
+      default:
+        CASSERT(false);
+    }
 
   mDaten[spezvar_kind] = s;
+  mDaten[spezvar_version] =
+    (v==-1 ? Aufnahme::rnd(ld->mSorten[s]->getVersions()) : v);
   
   /* In Zukunft soll das mal nicht mehr mglich sein:
   if (gGleichZeit) {
-    fprintf(stderr, "Neuer Blop in Gleichzeit\n");
+    print_to_stderr("Neuer Blop in Gleichzeit\n");
   }
   */
   
-  setWillInitEvent(true);
+  scheduleEvent(event_init);
 }
 
 
 
-
 Blop::Blop(const Blop & b): mFreilaufend(true), mBesitzer(0),
+  mOrt(absort_nirgends),
   mDatenLaenge(ld->mLevelKnoten->getDatenLaenge()),
   mZeitNummerDatenAlt(-1),
-  mZeitNummerInitStapel(-1),
-  mWillInitEvent(false)
+  mZeitNummerLeereStapel(-1),
+  mScheduleEventNr(event_keins)
 {
   CASSERT(b.mDaten);
   CASSERT(b.mDatenLaenge == ld->mLevelKnoten->getDatenLaenge());
@@ -96,7 +108,7 @@ Blop::Blop(const Blop & b): mFreilaufend(true), mBesitzer(0),
 }
 
 Blop::~Blop() {
-  setWillInitEvent(false);
+  unscheduleEvent();
   if (mDaten)
     delete[] mDaten;
   if (mDatenAlt)
@@ -104,10 +116,14 @@ Blop::~Blop() {
 }
 
 
-/** Tut, was man erwartet. */
+/** Tut, was man erwartet. Tut *nicht* fr reine Bildstapelblops. */
 Blop & Blop::operator=(const Blop & b) {
   CASSERT(b.mDaten);
   CASSERT(b.mDatenLaenge == ld->mLevelKnoten->getDatenLaenge());
+  
+  /* Innerhalb einer Gleichzeit Blops rumschieben fhrt zu Bugs
+     in Zukunfts-Zuweisungen... */
+  CASSERT(!gGleichZeit);
     
   if (!mDaten || mDatenLaenge != ld->mLevelKnoten->getDatenLaenge()) {
     /* Wir wechseln wohl grad in einen neuen Level. Zumindest
@@ -142,26 +158,44 @@ void Blop::kopiere(const Blop & b) {
       mDatenAlt[i] = b.mDatenAlt[i];
   }
   
-  mZeitNummerInitStapel = b.mZeitNummerInitStapel;
-  if (mZeitNummerInitStapel < gAktuelleZeitNummerInitStapel) {
+  mZeitNummerLeereStapel = b.mZeitNummerLeereStapel;
+  if (mZeitNummerLeereStapel < gAktuelleZeitNummerLeereStapel) {
     /* Bildstapel kopieren; aber nur, wenn er nicht sowieso
        frisch initialisiert werden muss. */
     mBild = b.mBild;
   }
-
-  /* BildAlt sollte nach einem Kopieren nicht mehr bentigt werden;
-     also brauchen wir es nicht zu kopieren. Aber vielleicht ist
-     unser mBildAlt noch von einem vorigen Level, und die Stapelgre
-     stimmt noch nicht. Deshalb neuen Bildstapel erzeugen.
-     Der neue Stapel hat brigens mAnz = -1, so dass - falls
+  
+  /* BildAlt sollte nach einem Kopieren nicht mehr bentigt werden; 	 
+     also brauchen wir es nicht zu kopieren. Aber vielleicht ist 	 
+     unser mBildAlt noch von einem vorigen Level, und die Stapelgre 	 
+     stimmt noch nicht. Deshalb neuen Bildstapel erzeugen. 	 
+     Der neue Stapel hat brigens mAnz = -1, so dass - falls 	 
      ber neumalen entschieden wird - auf jeden Fall neu gemalt wird. */
-  mBildAlt = BildStapel();
+  mBildAlt = BildStapel(); 	 
 
 
   mKettenGroesse = b.mKettenGroesse;
-  mAmPlatzen = b.mAmPlatzen;
 
-  setWillInitEvent(b.mWillInitEvent);
+  
+  unscheduleEvent();
+  if (b.mScheduleEventNr != event_keins)
+    scheduleEvent(b.mScheduleEventNr);
+}
+
+
+/** Verwandelt einen uninitialisierten Blop in einen Bildstapelblop;
+ *  oder erzeugt den bildstapel neu (fr Stapelhhennderungen bei
+ *  Levelwechsel */
+void Blop::erzeugeBildstapel() {
+  CASSERT(!mDaten);
+  
+  mBild = BildStapel();
+  mBildAlt = BildStapel();
+  /* mBild muss noch "leer" gemacht werden. (im Moment hat es noch -1
+   * bilder) */
+  mZeitNummerLeereStapel = -1;
+  /* Der Ort ist ein Dummywert; nur dazu da, um zu erlauben, hier hinzumalen */
+  mOrt = ort_absolut(absort_bemalbar);
 }
 
 
@@ -169,46 +203,33 @@ void Blop::kopiere(const Blop & b) {
 /** Setzt Besitzer und Besitzer-Informationen. Braucht nur am Anfang einmal
     aufgerufen zu werden. Muss auch fr den Global-Blop aufgerufen werden;
     sonst darf kein Code ausgefhrt werden. */
-void Blop::setBesitzer(BlopBesitzer * bes /*= 0*/, bool re /*=false*/,
-		       int x /*= keine_koord*/,
-		       int y /*= keine_koord*/) {
+void Blop::setBesitzer(BlopBesitzer * bes /*= 0*/,
+    ort_absolut ort /* = ort_absolut(absort_nirgends)*/) {
   mFreilaufend = false;
   mBesitzer = bes;
-  mRechterSpieler = re;
-  mX = x;
-  mY = y;
+  mOrt = ort;
 }
 
 
-/** Setzt die Version-Spezvar des Blops. Wird fr Grasblops am Anfang
-    aufgerufen. */
-void Blop::setVersion(int v) {
-  mDaten[spezvar_version] = v;
-}
-
-
-
-
 
 
 /** malt den Blop; xx und yy sind in Pixeln angegeben;
     der Hintergrund wird vorher gelscht. */
-void Blop::malen(QPainter & p, int xx, int yy) const {
+void Blop::malen(int xx, int yy, bool apply_mirror /*=true*/) const {
   /* Ggf. gelazyten initStapel()-Aufruf nachholen */
-  braucheInitStapel();
+  braucheLeereStapel();
 
   try { // Um die Fehlermeldung zu verbessern...
-
-    CASSERT(mDaten);
     
     //if (mX == 0 && mY == 19)
     //  mBild.print();
 
     /* Der Bildstapel soll sich geflligst selbst malen */
-    mBild.malen(p, xx, yy);
+    mBild.malen(xx, yy, apply_mirror);
 
   } catch (Fehler fe) {
-    throw Fehler(toString() + ", during drawing:\n" + fe.getText());
+    throw Fehler("%s, during drawing:\n%s",
+		 toString().data(),fe.getText().data());
   }
 
 } // malen
@@ -249,7 +270,7 @@ int Blop::getVersion() const {
 /** liefert true, wenn der Blop am platzen ist */
 bool Blop::getAmPlatzen() const {
   CASSERT(mDaten);
-  return mAmPlatzen != 0;
+  return mDaten[spezvar_am_platzen] != 0;
 }
 
 
@@ -260,15 +281,14 @@ bool Blop::getAmPlatzen() const {
 void Blop::animiere() {
   CASSERT(gGleichZeit);
   CASSERT(!mFreilaufend);
-  /* Wenn ein Blop auf ein Init-Event wartet, darf noch nix anderes
-     aufgerufen werden. */
-  CASSERT(!mWillInitEvent);
+  /* Wenn animiert wird, sollten keine anderen Events ausstehen. */
+  CASSERT(mScheduleEventNr == event_keins);
 
   /* Ggf. gelazyten initStapel()-Aufruf nachholen. Man knnte genau
      untersuchen, wann whrend animiere() dieser Aufruf bentigt wird,
      aber irgend wann wird er sowieso in jedem Spielschritt bentigt,
      also macht es keinen Sinn, sich Mhe zu geben. */
-  braucheInitStapel();
+  braucheLeereStapel();
   
   try { // Um ggf. die Fehlermeldung zu verbessern...
 
@@ -287,9 +307,11 @@ void Blop::animiere() {
        verschicken. */
     if (mDaten[spezvar_kind] != mDaten[spezvar_kind_beim_letzten_draw_aufruf]
         && mDaten[spezvar_kind_beim_letzten_draw_aufruf] != -1) {
-       Code * alt_co = ld->mSorten[
-                           mDaten[spezvar_kind_beim_letzten_draw_aufruf]
-			 ]->getEventCode(event_draw);
+      Code * alt_co =
+        (mDaten[spezvar_kind_beim_letzten_draw_aufruf] == blopart_ausserhalb
+          ? 0
+          : ld->mSorten[mDaten[spezvar_kind_beim_letzten_draw_aufruf]]
+	      ->getEventCode(event_draw));
        if (alt_co)
          alt_co->busyReset(*this);
     }
@@ -302,7 +324,7 @@ void Blop::animiere() {
     if (mc) {    
       /* Bei diesem eval werden die Bildchen neu in mBild eingefgt...
          Das ist beim Global-Blop aber nicht erlaubt. */
-      mMalenErlaubt = mDaten[spezvar_kind] != blopart_global;
+      mMalenErlaubt = true;
       mc->eval(*this);
     }
 
@@ -318,45 +340,27 @@ void Blop::animiere() {
        Cual-Programms mal ganz kurz das leere Blop ist, dann normalerweise
        so kurz, dass die nachfolgenden Zeilen es nicht merken. */
     if (mDaten[spezvar_kind] == blopart_keins)
-      mAmPlatzen = 0;
+      mDaten[spezvar_am_platzen] = 0;
     
-    if (mAmPlatzen > 0) {
-      mAmPlatzen++;
-      if (mAmPlatzen > platzbild_anzahl) {
+    if (mDaten[spezvar_am_platzen] > 0) {
+      mDaten[spezvar_am_platzen]++;
+      if (mDaten[spezvar_am_platzen] > ld->mExplosionBild.anzBildchen()) {
         /* Letztes Platz-Bildchen. Im nchsten Schritt in
            nichts verwandeln. */
-        mDaten[spezvar_kind] = blopart_keins;
+        setKindIntern(blopart_keins);
 	/* Wenn der Blop sich als Gras verhlt, dann soll er das nicht
 	   mehr tun, nachdem er geplatzt ist. */
-	mDaten[spezvar_goalblob] = 0;
-	mAmPlatzen = 0;
+	mDaten[spezvar_am_platzen] = 0;
       }
     } // if Blop ist am platzen
     
   } catch (Fehler fe) {
-    throw Fehler(toString() + ", during animation:\n" + fe.getText());
+    throw Fehler("%s, during animation:\n%s",
+                 toString().data(), fe.getText().data());
   }
 }
 
 
- 
-/** Liefert true, wenn der Blop sich seit dem letzten Aufruf von
-    takeUpdaten() verndert hat und deshalb neu gezeichnet werden muss.
-    Liefert auf jeden Fall true, wenn der Blop zwischendrin kopiert wurde
-    (mit = oder copy-Constructor). */
-bool Blop::takeUpdaten() {
-  CASSERT(mDaten);
-
-  /* Ggf. gelazyten initStapel()-Aufruf nachholen */
-  braucheInitStapel();
-
-  if (mBild == mBildAlt)
-    return false;
-  mBildAlt = mBild;
-  return true;
-}
-
-
 
 /** Fhrt den Code des angegebenen Events aus (falls Code zu diesem
     Event existiert. Sollte innerhalb einer Gleichzeit aufgerufen
@@ -369,10 +373,9 @@ void Blop::execEvent(int evtnr) {
     CASSERT(gGleichZeit);
 
     CASSERT(!mFreilaufend);
-    /* Wenn ein Blop auf ein Init-Event wartet, darf noch nix anderes
-       aufgerufen werden. (mWillInitEvent wird _vor_ der Event-Versendung
-       auf false gesetzt.) */
-    CASSERT(!mWillInitEvent);
+    /* Wenn ein Event aufgerufen wird, sollte kein anderes Event ausstehen.
+       (mScheduleEventNr wird gelscht, _bevor_ das entsprechende Event gesendet wird.) */
+    CASSERT(mScheduleEventNr == event_keins);
 
     CASSERT(mDaten);
 
@@ -384,7 +387,7 @@ void Blop::execEvent(int evtnr) {
     }
 
   } catch (Fehler fe) {
-    throw Fehler(_("%s, during %s event:\n%s"),
+    throw Fehler("%s, during %s event:\n%s",
               toString().data(), cEventNamen[evtnr], fe.getText().data());
   }
 }
@@ -396,16 +399,15 @@ void Blop::execEvent(int evtnr) {
 void Blop::lassPlatzen() {
   CASSERT(mDaten);
   if (mDaten[spezvar_kind] == blopart_keins)
-    throw Fehler(_("Empty Blops are not allowed to explode."));
-  if (mAmPlatzen == 0)
-    mAmPlatzen = 1;
+    throw Fehler("%s","Empty Blops are not allowed to explode.");
+  if (mDaten[spezvar_am_platzen] == 0)
+    mDaten[spezvar_am_platzen] = 1;
 }
 
 
 /** Teilt einem Farbblop die (neue) Gre seiner Kette mit. */
 void Blop::setKettenGroesse(int anz) {
   CASSERT(mDaten);
-  CASSERT(mDaten[spezvar_kind] >= 0);
 
   /* Kettengre abspeichern, fr den Fall, dass sie als spezconst
      abgefragt wird */
@@ -414,6 +416,22 @@ void Blop::setKettenGroesse(int anz) {
 
 
 
+/** Liefert true, wenn der Blop sich seit dem letzten Aufruf von 	 
+    takeUpdaten() verndert hat und deshalb neu gezeichnet werden muss.
+    Liefert auf jeden Fall true, wenn der Blop zwischendrin kopiert wurde 
+    (mit = oder copy-Constructor). */ 	 
+bool Blop::takeUpdaten() { 	 
+  	 
+  /* Ggf. gelazyten initStapel()-Aufruf nachholen */ 	 
+  braucheLeereStapel(); 	 
+  	 
+  if (mBild == mBildAlt) 	 
+    return false; 	 
+  mBildAlt = mBild; 	 
+  return true; 	 
+}
+
+
 
 /***** Funktionen, die nix verndern *****/
 
@@ -431,8 +449,10 @@ bool Blop::verbindetMit(const Blop & b) const {
 
 
 /** Liest die aktuelle Sorte dieses Blops aus den Leveldaten aus. */
-Sorte * Blop::getSorte() const {
-  return ld->mSorten[mDaten[spezvar_kind]];
+Sorte * Blop::getSorte(bool vergangenheit /*= false*/) const {
+  return ld->mSorten[vergangenheit
+		     ? getVariableVergangenheit(spezvar_kind)
+		     : getVariable(spezvar_kind)];
 }
 
 /** Fragt beim Besitzer an, in welche Richtungen dieser Blop
@@ -440,7 +460,7 @@ Sorte * Blop::getSorte() const {
 int Blop::getVerbindungen() const {
   CASSERT(mDaten);
   if (mBesitzer) {
-    return mBesitzer->getBesitzVerbindungen(mX, mY);
+    return mBesitzer->getBesitzVerbindungen(mOrt.x, mOrt.y);
   } else
     return verbindung_solo;
 }
@@ -456,41 +476,59 @@ bool Blop::verbindetMitRand(int seite) const {
     beitrgt (normalerweise 1). */
 int Blop::getKettenBeitrag() const {
   CASSERT(mDaten);
-  CASSERT(mDaten[spezvar_kind] >= 0);
-  if (getSorte()->getKetteZaehltNurInnen()) {
-    return (getVerbindungen() & verbindung_alle4) == verbindung_alle4;
-  } else
-    return 1;
+  return mDaten[spezvar_gewicht];
 }
 
+/** Liefert zurck, ob ein bestimmtes Platzverhalten vorliegt */
+bool Blop::getVerhalten(int verhalten) const {
+  return (mDaten[spezvar_verhalten] & verhalten)==verhalten;
+}
+
+
+/** Liefert zurck, welchem Spieler (0 oder 1) der Blop gehoert oder
+    2 bei globalblop */
+int Blop::getSpieler() const {
+  switch (mOrt.art) {
+    case absort_feld:
+    case absort_fall:
+    case absort_semiglobal:
+    case absort_info:
+      return mOrt.rechts;
+    case absort_global:
+    case absort_nirgends:
+      return 2;
+  }
+  CASSERT(0);
+  return 0;
+}
+
+
 
 /** Liefert einen String der Art "Blop Drachen at x,y".
     Fr Fehlermeldungen. */
-__String Blop::toString() const {
-  __String ret1, ret2;
-  ret1.sprintf(_("Blop %s"), getSorte()->getName().data());
-
-  if (mY == keine_koord) {
-    ret2.sprintf(_("%s"), ret1.data());
-  } else {
-    ret2.sprintf(_("%s at %d,%d"), ret1.data(), mX, mY);
-  }
-  return ret2;
+Str Blop::toString() const {
+  return _sprintf("Blop %s at %s", getSorte()->getName().data(),
+    mOrt.toString().data());
 }
 
 
 /** Wird vom Code aufgerufen, wenn es Punkte geben soll. */
 void Blop::bekommPunkte(int pt) const {
   if (getArt() == blopart_global)
-    throw Fehler(_("bonus() does not work in the global blob."));
-  Cuyo::gCuyo->neuePunkte(mRechterSpieler, pt);
+    throw Fehler("%s","bonus() does not work in the global blob.");
+  Cuyo::neuePunkte(mOrt.rechts, pt);
 }
 
 /** Wird vom Code aufgerufen, wenn ein Message angezeigt werden soll. */
-void Blop::zeigMessage(__String mess) const {
+void Blop::zeigMessage(Str mess) const {
   if (getArt() == blopart_global)
-    throw Fehler(_("message() does not work in the global blob."));
-  Cuyo::gCuyo->getSpielfeld(mRechterSpieler)->setMessage(mess);
+    throw Fehler("%s","message() does not work in the global blob.");
+  Cuyo::getSpielfeld(mOrt.rechts)->setMessage(mess);
+}
+
+
+const ort_absolut & Blop::getOrt() const {
+  return mOrt;
 }
 
 
@@ -520,81 +558,65 @@ void Blop::initSchritt() {
 extern int spezconst_default[spezconst_anz];
 
 
-int Blop::getSpezConst(int vnr) const {
+int Blop::getSpezConst(int vnr, bool vergangenheit /*= false*/) const {
   CASSERT(mDaten);
   switch (vnr) {
     case spezconst_connect:
       return getVerbindungen();
-     case spezconst_size:
-       return mKettenGroesse;
-     case spezconst_loc_x:
-       /* Wenn die *Y*-Koordinate keine_koord ist, gehre ich
-          vermutlich dem Fall. */
-       if (mY == keine_koord) break;
-       return ld->mSpiegeln ? grx - 1 - mX : mX;
-     case spezconst_loc_y:
-       if (mY == keine_koord) break;
-       return ld->mSpiegeln ? gry - 1 - mY : mY;
-     case spezconst_loc_p:
+    case spezconst_size:
+      return mKettenGroesse;
+    case spezconst_loc_x:
+      if (mOrt.art != absort_feld) break;
+      return ld->mSpiegeln ? grx - 1 - mOrt.x : mOrt.x;
+    case spezconst_loc_y:
+      if (mOrt.art != absort_feld) break;
+      return ld->mSpiegeln ? gry - 1 - mOrt.y : mOrt.y;
+    case spezconst_loc_p:
+      if ((mOrt.art == absort_global) || (mOrt.art == absort_nirgends))
+        throw iFehler(
+        "Internal error in Blop::getSpezConst(): Wrong AbsOrtArt %d",
+        mOrt.art);
        /* Hier sollte noch eine Fehlermeldung ausgegeben werden, wenn man
           das vom Global-Blop aus abfragen will. */
-       return mRechterSpieler ? 2 : 1;
-     case spezconst_players:
-       return ld->mSpielerZahl;
-     case spezconst_exploding:
-       return mAmPlatzen;
-  };
+      return mOrt.rechts ? 2 : 1;
+    case spezconst_players:
+      return ld->mSpielerZahl;
+    case spezconst_exploding:
+      return getVariableVergangenheit(spezvar_am_platzen);
+    case spezconst_basekind:
+      return getSorte(vergangenheit)->getBasekind();
+    case spezconst_time:
+      return Cuyo::getSpielfeld(false)-> getZeit();
+    case spezconst_info:
+      return mOrt.art==absort_info || (mOrt.art==absort_fall && mOrt.y);
+  }
 
   /* Nicht selbst bearbeitet? Dann nach oben weitergeben. */
   if (mBesitzer) {
-    
-    /* Wenn der Besitzer der Fall ist, will der wissen, welcher der
-       beiden Blops ich bin... */
-    if (mY == keine_koord && mX == 1)
-      vnr += spezconst_bin_1;
-    
-    int r = mBesitzer->getSpezConst(vnr);
+    int r = mBesitzer->getSpezConst(vnr,this);
     if (r != spezconst_defaultwert)
       return r;
   }
 
   /* Wir haben keinen Besitzer oder der Besitzer weigert sich, uns 
      was mitzuteilen? Dann Default-Wert. */
-  return spezconst_default[vnr];
+  return spezconst_default[-vnr-1];
 }
 
 
 /** High-Level: Wird benutzt, wenn eine Variable im cual-Programm steht.
     Kmmert sich auch um all das @()-Zeug und die Zeitverzgerung. */
-int Blop::getVariable(const Variable & v) const {
-  switch (v.getOrt()) {
-  case variable_lokal:
+int Blop::getVariable(const Variable & v) {
+  if (v.Ort_hier())
     return getVariable(v.getNummer());
-    
-  case variable_global:
-    return gGlobalBlop.getVariableVergangenheit(v.getNummer());
-    
-  case variable_relativ: {
-    int xx = v.getX(mX, mY);
-    int yy = v.getY(mX, mY);
-
-    if (mBesitzer && mBesitzer->koordOK(xx, yy))
-      return mBesitzer->getFeld(xx, yy).
-               getVariableVergangenheit(v.getNummer());
-    else {
-      /* An den angegebenen Koordinaten relativ zu diesem Blop
-         gibt's grad nix. Dann default zurckliefern. */
+  else {
+    ort_absolut ziel = v.getOrt(mOrt, *this);
+    if (ziel.korrekt())
+      return ziel.finde().getVariableVergangenheit(v.getNummer());
+    else  /* Es ist ein outside-Blob,
+             also den Default unabhngig von der Default-Art nehmen. */
       return v.getDefaultWert();
-    }
   }
-
-  case variable_fall:
-    return mBesitzer->getFall(v.getWelcherFall())->getVariable(v.getNummer());
-
-  default:
-    CASSERT(false);
-  }
-  return 0; // nur, um keine Warnung zu kriegen.
 }
 
 
@@ -606,42 +628,15 @@ int Blop::getVariable(const Variable & v) const {
 void Blop::setVariable(const Variable & v, int wert, int op) {
   /* Da will doch nicht jemand etwa eine Spezconst ndern?! */
   if (v.getNummer() < 0)
-    throw Fehler(_("Variable \"%s\" is read-only."),
+    throw Fehler("The variable \"%s\" is read-only.",
                  v.toString().data());
-  /* Range-Check bei Spez-Vars */
-  /* file, pos und qu werden erst beim Malen gecheckt. version braucht
-     gar nicht gecheckt zu werden. out1 und out2 auch nicht. */
-  switch (v.getNummer()) {
-    case spezvar_kind:
-      if (wert < blopart_min_cual || wert >= ld->mAnzFarben)
-        throw Fehler(_("Variable kind (=%d) out of range (allowed: %d - %d)"),
-	             wert, blopart_min_cual, ld->mAnzFarben - 1);
-      break;
-  }
-  
-  
-  switch (v.getOrt()) {
-  case variable_lokal:
+
+  if (v.Ort_hier())
     setVariable(v.getNummer(), wert, op);
-    break;
-  case variable_global:
-    gGlobalBlop.setVariableZukunft(v.getNummer(), wert, op);
-    break;
-  case variable_relativ: {
-    int xx = v.getX(mX, mY);
-    int yy = v.getY(mX, mY);
-    if (mBesitzer && mBesitzer->koordOK(xx, yy)) {
-      mBesitzer->getFeld(xx, yy).
-        setVariableZukunft(v.getNummer(), wert, op);
-    }
-    break;
-  }
-  case variable_fall:
-    mBesitzer->getFall(v.getWelcherFall())
-      ->setVariable(v.getNummer(), wert, op);
-    break;
-  default:
-    CASSERT(false);
+  else {
+    ort_absolut ziel = v.getOrt(mOrt, *this);
+    if (ziel.korrekt())
+      ziel.finde().setVariableZukunft(v.getNummer(), wert, op);
   }
 }
 
@@ -675,18 +670,56 @@ void Blop::setVariable(int vnr, int wert, int op) {
 void Blop::setVariableIntern(int vnr, int wert, int op) {
   CASSERT(mDatenLaenge == ld->mLevelKnoten->getDatenLaenge());
   CASSERT(vnr >= 0 && vnr < ld->mLevelKnoten->getDatenLaenge());
-  
+
+  int oldValue = mDaten[vnr];
+
   switch (op) {
     case set_code: mDaten[vnr] = wert; break;
     case add_code: mDaten[vnr] += wert; break;
     case sub_code: mDaten[vnr] -= wert; break;
     case mul_code: mDaten[vnr] *= wert; break;
-    case div_code: mDaten[vnr] /= wert; break;
-    case mod_code: mDaten[vnr] %= wert; break;
+    case div_code: mDaten[vnr] = divv(mDaten[vnr],wert); break;
+    case mod_code: mDaten[vnr] = modd(mDaten[vnr],wert); break;
+    case bitset_code: mDaten [vnr] |= wert; break;
+    case bitunset_code: mDaten [vnr] &= (-1-wert); break;
     default:
-      throw iFehler(_("Internal error in Blop::setVariable() (op = %d)"),
+      throw iFehler("Internal error in Blop::setVariable() (op = %d)",
                    op);
   }
+
+  /* Range-Check bei Spez-Vars */
+  /* file, pos und qu werden erst beim Malen gecheckt. version braucht
+     gar nicht gecheckt zu werden. out1 und out2 auch nicht. */
+  switch (vnr) {
+    case spezvar_kind:
+      if (mDaten[vnr] < blopart_min_cual || mDaten[vnr] >= ld->mAnzFarben)
+        throw Fehler("Value %d for kind out of range (allowed: %d - %d)",
+	             mDaten[vnr], blopart_min_cual, ld->mAnzFarben - 1);
+      break;
+  }
+
+  if ((vnr==spezvar_kind) && (mDaten[vnr]!=oldValue))
+    setKindIntern(mDaten[vnr]);
+}
+
+
+/** spezvar_kind ist eine ganz spezielle spezvar: Ganz lowlevel passiert
+    hier noch mehr. */
+void Blop::setKindIntern(int wert) {
+  mDaten[spezvar_kind] = wert;
+  for (int i = 0; i < mDatenLaenge; i++)
+    switch (ld->mSorten[wert]->getDefaultArt(i)) {
+      case da_nie:
+      case da_keinblob:
+      case da_init:
+      case da_event:
+	break;
+      case da_kind:
+        mDaten[i] = ld->mSorten[wert]->getDefault(i);
+	break;
+      default:
+	CASSERT(false);
+    }
 }
 
 
@@ -702,7 +735,7 @@ int Blop::getVariableVergangenheit(int vnr) const {
   CASSERT(mDatenLaenge == ld->mLevelKnoten->getDatenLaenge());
   
   if (vnr < 0)
-    return getSpezConst(vnr);
+    return getSpezConst(vnr,true);
   
   CASSERT(vnr < ld->mLevelKnoten->getDatenLaenge());
   
@@ -750,13 +783,14 @@ void Blop::merkeAlteVarWerte() {
 
 
 /** Schaut, ob noch ein Lazy-Evaluation-initStapel()-Aufruf aussteht
-    und fhrt ihn ggf. aus. (Siehe lazyInitStapel().)
+    und fhrt ihn ggf. aus. (Siehe lazyLeereStapel().)
     Ist zwar eigentlich nicht wirklich const, wird aber in Situationen
     aufgerufen, wo man sich eigentlich const fhlen mchte. */
-void Blop::braucheInitStapel() const {
-  if (mZeitNummerInitStapel < gAktuelleZeitNummerInitStapel) {
-    ((BildStapel &) mBild).initStapel(mAmPlatzen);
-    (long &) mZeitNummerInitStapel = gAktuelleZeitNummerInitStapel;
+void Blop::braucheLeereStapel() const {
+  CASSERT(mBild.istInitialisiert());
+  if (mZeitNummerLeereStapel < gAktuelleZeitNummerLeereStapel) {
+    ((BildStapel &) mBild).initStapel(mDaten ? mDaten[spezvar_am_platzen] : 0);
+    (long &) mZeitNummerLeereStapel = gAktuelleZeitNummerLeereStapel;
   }
 }
 
@@ -770,88 +804,114 @@ void Blop::speichereBild() {
   /* Ggf. gelazyten initStapel()-Aufruf nachholen. Eigentlich unntig,
      da das schon am Anfang von animiere() passiert sein sollte. Ich
      fhl' mich aber wohler. */
-  braucheInitStapel();
+  braucheLeereStapel();
 
-  if (!mMalenErlaubt)
-    throw Fehler(_("Drawing is not allowed at the moment."));
+  if ((!mMalenErlaubt) || (!mOrt.bemalbar()))
+    throw Fehler("%s","Drawing is not allowed at the moment.");
   mBild.speichereBild(getSorte(), mDaten[spezvar_file], mDaten[spezvar_pos],
      mDaten[spezvar_quarter]);
 }
 
+
 /** Speichert das aktuelle Bild (d. h. aus den spezvar file und pos)
     in die Mal-Liste von einem anderen Blop, und zwar so, dass es
-    erst zum Schluss gemalt wird. */
-void Blop::speichereBildRelativ(int xx, int yy) {
+    in Ebene ebene gemalt wird. */
+void Blop::speichereBildFremd(Ort & ort, int ebene) {
   CASSERT(mDaten);
-  if (!mMalenErlaubt)
-    throw Fehler(_("Drawing is not allowed at the moment."));
 
-  /* Die Koordinatenberechnung ist aus Variable::getX / getY abgeschrieben
-     und sollte irgendwo zentral sein. */
-  int x = mX + xx;
-  int y = mY + (ld->mSpiegeln ? -yy : yy)
-	              - (ld->mSechseck && (mX & 1) && (xx & 1));
+  ort_absolut ziel = ort.berechne(mOrt,*this);
+  if (ziel.korrekt(true)) {
+    Blop & b = ziel.finde();
+
+    if ((!mMalenErlaubt) || (!b.mOrt.bemalbar()))
+      throw Fehler("%s","Drawing is not allowed at the moment.");
 
-  if (mBesitzer && mBesitzer->koordOK(x, y)) {
-    Blop & b = mBesitzer->getFeld(x, y);
     /* Ggf. gelazyten initStapel()-Aufruf nachholen. An dieser Stelle
        ist das wichtig, weil der bemalte Blop evtl. noch nicht animiert()
        wurde. */
-    b.braucheInitStapel();
+    b.braucheLeereStapel();
 
-    /* letztes true sagt, dass das Bild erst am Ende gemalt werden soll. */
     b.mBild.speichereBild(
           getSorte(), mDaten[spezvar_file], mDaten[spezvar_pos],
-          mDaten[spezvar_quarter], true);
+          mDaten[spezvar_quarter], ebene);
   }
 }
 
 
 
-/********** Init-Event-Verwaltung **********/
+/********** Scheduled-Event-Verwaltung **********/
 
-/** Setzt, ob dieser Blop ein Init-Event will. */
-void Blop::setWillInitEvent(bool wie) {
-  if (mWillInitEvent == wie)
-    return;
-  mWillInitEvent = wie;
+
+/** Wie execEvent; wird aber nicht sofort ausgefhrt, sondern erst
+    beim nchsten Aufruf von sendeGeschedulteEvents() */
+void Blop::scheduleEvent(int evtnr) {
+  /* Bisher ist es noch nicht mglich, mehrere Events fr einen Blop
+   * zu schedulen. */
+  CASSERT(mScheduleEventNr == event_keins);
   
-  if (mWillInitEvent) {
-    mWIEPos = gWIEListe.size();
-    gWIEListe.resize(mWIEPos + 1);
-    gWIEListe[mWIEPos] = this;
-    
-  } else {
-    gWIEListe[mWIEPos] = 0;
+  /* Eigentlich gibt es keinen Grund, innerhalb einer Gleichzeit zu verbieten, ein
+   * Event fr spter zu schedulen. Bisher (2014) sollte das aber nicht vorkommen, und
+   * wenn doch, dann heit das, dass irgend was anderes falsch ist. */
+  CASSERT(!gGleichZeit);
+
+  mScheduleEventNr = evtnr;
+  mSEPos = gSEListe.size();
+  gSEListe.push_back(this);
+}
+
+
+/** Lscht diesen Blop aus gWEListe (falls ntig) */
+void Blop::unscheduleEvent() {
+  if (mScheduleEventNr != event_keins) {
+    gSEListe[mSEPos] = 0;
+    mScheduleEventNr = event_keins;
   }
 }
 
 
-QArray<Blop *> Blop::gWIEListe;
 
-/** Ruft alle noch fehlenden Init-Events aus. */
-void Blop::sendeFehlendeInitEvents() {
+std::vector<Blop *> Blop::gSEListe;
+
+/** Ruft alle geschedulten Events aus. */
+void Blop::sendeGeschedulteEvents() {
+  /* Um mglichst kurze gleichzeiten zu haben, knnte man jede Event-Sorte in einer
+     eigenen Gleichzeit ausfhren. Bisher wei ich aber noch nicht, in welcher Reihenfolge
+     ich das machen wollen wrde. Das einzige, was wirklich logisch erscheint, ist, init-Events
+     vor allen anderen auszufhren. */
+  
   beginGleichzeitig();
-  for (int i = 0; i < (int) gWIEListe.size(); i++)
-    if (gWIEListe[i]) {
-      gWIEListe[i]->mWillInitEvent = false;
-      gWIEListe[i]->execEvent(event_init);
+  for (int i = 0; i < (int) gSEListe.size(); i++)
+    if (gSEListe[i] && gSEListe[i]->mScheduleEventNr == event_init) {
+      gSEListe[i]->mScheduleEventNr = event_keins;
+      gSEListe[i]->execEvent(event_init);
+      gSEListe[i] = 0;
     }
-  gWIEListe.resize(0);
   endGleichzeitig();
+
+  beginGleichzeitig();
+  for (int i = 0; i < (int) gSEListe.size(); i++)
+    if (gSEListe[i]) {
+      int e = gSEListe[i]->mScheduleEventNr;
+      gSEListe[i]->mScheduleEventNr = event_keins;
+      gSEListe[i]->execEvent(e);
+    }
+  endGleichzeitig();
+  gSEListe.clear();
 }
 
 
+void Blop::playSample(int nr) const {mOrt.playSample(nr);}
+
 
 
 /********** Statisches Zeug **********/
 
 bool Blop::gGleichZeit = false;
 long Blop::gAktuelleZeitNummerDatenAlt = 0;
-long Blop::gAktuelleZeitNummerInitStapel = 0;
+long Blop::gAktuelleZeitNummerLeereStapel = 0;
 /** Liste Zukunfts-Zuweisungen, die sich whrend einer Gleichzeit
     ansammeln. */
-QArray<Blop::tZZ> Blop::gZZ;
+std::vector<Blop::tZZ> Blop::gZZ;
 int Blop::gZZAnz;
 //bool Blop::gInitEventsAutomatisch = true;
 
@@ -899,11 +959,11 @@ void Blop::abbruchGleichzeitig() {
     d. h. erst dann, wenn's wirklich gebraucht wird.
     Gebraucht wird's natrlich, wenn ein Blop animiert wird. Aber auch,
     wenn ein Nachbarblop etwas auf diesen Blop malt. */
-void Blop::lazyInitStapel() {
-  /* Wenn ein Blop sieht, dass seine mZeitNummerInitStapel-Variable
-     kleiner ist als gAktuelleZeitNummerInitStapel, ruft er initStapel()
+void Blop::lazyLeereStapel() {
+  /* Wenn ein Blop sieht, dass seine mZeitNummerLeereStapel-Variable
+     kleiner ist als gAktuelleZeitNummerLeereStapel, ruft er initStapel()
      auf. */
-  gAktuelleZeitNummerInitStapel++;
+  gAktuelleZeitNummerLeereStapel++;
 }
 
 
