@@ -4,6 +4,9 @@
     begin                : Thu Jul 20 2000
     copyright            : (C) 2000 by Immi
     email                : cuyo@pcpool.mathematik.uni-freiburg.de
+
+Modified 2001-2003,2005,2006,2008,2010,2011,2014 by the cuyo developers
+
  ***************************************************************************/
 
 /***************************************************************************
@@ -18,19 +21,22 @@
 #ifndef BLOP_H
 #define BLOP_H
 
-#include "inkompatibel.h"
+#include <vector>
 
-#include <qpainter.h>
-#include <qstring.h>
+#include <SDL.h>
 
-#include "leveldaten.h"
-#include "blopbesitzer.h"
 #include "bildstapel.h"
+#include "ort.h"
+
+class Str;
+class BlopBesitzer;
+class Variable;
+class VarDefinition;
 
 /* blopart_... wird jetzt in sorte.h definiert... */
 
 // liefert getFarbe manchmal zurck
-#define keine_farbe -1
+#define keine_farbe (-1)
 
 // Wert von mX und mY, falls das Blop nicht im Gitter ist
 #define keine_koord 0x7fff
@@ -48,8 +54,12 @@
 #define spezvar_out2 6
 #define spezvar_kind_beim_letzten_draw_aufruf 7
 #define spezvar_inhibit 8
-#define spezvar_goalblob 9
-#define spezvar_anz 10
+#define spezvar_gewicht 9
+#define spezvar_verhalten 10
+#define spezvar_falling_speed 11
+#define spezvar_falling_fast_speed 12
+#define spezvar_am_platzen 13
+#define spezvar_anz 14
 /* spezvar_namen[] in knoten.cpp definiert. Es enthlt auch
    die Namen der spezconst */
 
@@ -66,12 +76,10 @@
 #define spezconst_exploding (-10)
 #define spezconst_loc_xx (-11)
 #define spezconst_loc_yy (-12)
-#define spezconst_anz 12
-
-/* Nur zur bergabe in getSpezConst() vom Blop an den Fall-Besitzer.
-   Wird zu spezconst dazugezhlt, um mitzuteilen, dass ich Blop Nr. 1
-   bin. */
-#define spezconst_bin_1 -1000
+#define spezconst_basekind (-13)
+#define spezconst_time (-14)
+#define spezconst_info (-15)
+#define spezconst_anz 15
 
 
 /* Return-Wert von BlopBesitzer::getSpezConst(), der besagt, dass der
@@ -83,9 +91,14 @@
 #define bits_pro_int (8*sizeof(int))
 
 
+/* Bit-Konstanten fr spezvar_verhalten */
+#define platzt_bei_gewicht 1
+#define platzt_bei_platzen 2
+#define platzt_bei_kettenreaktion 4
+#define berechne_kettengroesse 8
+#define verhindert_gewinnen 16
+#define schwebt 32
 
-class Variable;
-class VarDefinition;
 
 
 
@@ -103,44 +116,41 @@ class Blop {
       was anderes nicht initialisiert ist). Das Uninitialisiert sein
       eines Blops erkennt man daran, dass mDaten 0 ist. */
   Blop();
-  /** Constructor... s ist Sorte. */
-  Blop(int s);
+  /** Constructor... s ist Sorte und v Version. v=-1 heit Zufallsversion. */
+  Blop(int s, int v=-1);
   /** Copy-Contructor... */
   Blop(const Blop & b);
   /** Destructor */
   ~Blop();
-  /** Tut, was man erwartet. */
+  /** Tut, was man erwartet. Tut *nicht* fr reine Bildstapelblops. */
   Blop & operator=(const Blop & b);
   /** kopiert die Daten von b. Erwartet, dass die Datenlngen schon
       bereinstimmen. */
   void kopiere(const Blop & b);
+  /** Verwandelt einen uninitialisierten Blop in einen Bildstapelblop;
+   *  oder erzeugt den bildstapel neu (fr Stapelhhennderungen bei
+   *  Levelwechsel */
+  void erzeugeBildstapel();
 
   /** Setzt Besitzer und Besitzer-Informationen. Braucht nur am Anfang einmal
       aufgerufen zu werden. Muss auch fr den Global-Blop aufgerufen werden;
       sonst darf kein Code ausgefhrt werden. */
-  void setBesitzer(BlopBesitzer * bes = 0, bool re = false,
-                   int x = keine_koord, int y = keine_koord);
-  /** Setzt die Version-Spezvar des Blops. Wird fr Grasblops am Anfang
-      aufgerufen. */
-  void setVersion(int v);
+  void setBesitzer(BlopBesitzer * bes = 0,
+    ort_absolut ort = ort_absolut(absort_nirgends));
 
 
   /** Fhrt die ganzen Animationen durch (Codeanimationen und platzen).
       Sollte innerhalb einer Gleichzeit aufgerufen werden. */
   void animiere();
-
-  /** Liefert true, wenn der Blop sich seit dem letzten Aufruf von
-      takeUpdaten() verndert hat und deshalb neu gezeichnet werden muss.
-      Liefert auf jeden Fall true, wenn der Blop zwischendrin kopiert wurde
-      (mit = oder copy-Constructor). */
-  bool takeUpdaten();
-
   
   /** Fhrt den Code des angegebenen Events aus (falls Code zu diesem
       Event existiert. Sollte innerhalb einer Gleichzeit aufgerufen
       werden.
       Die Event-Nummern sind in sorte.h definiert. */
   void execEvent(int evtnr);
+  /** Wie execEvent; wird aber nicht sofort ausgefhrt, sondern erst
+      beim nchsten Aufruf von sendeGeschedulteEvents() */
+  void scheduleEvent(int evtnr);
   /** Startet den Platzvorgang. Sollte nicht fr leere Blops aufgerufen
       werden. */
   void lassPlatzen();
@@ -150,13 +160,21 @@ class Blop {
   void setKettenGroesse(int anz);
 
 
+  /** Liefert true, wenn der Blop sich seit dem letzten Aufruf von 	 
+      takeUpdaten() verndert hat und deshalb neu gezeichnet werden muss.
+      Liefert auf jeden Fall true, wenn der Blop zwischendrin kopiert wurde
+      (mit = oder copy-Constructor). */ 	 
+  bool takeUpdaten();
+
 
   /***** Funktionen, die nix verndern *****/
 
   /** malt den Blop, wobei xx und yy in Pixeln
       angegeben ist; der Hintergrund wird vorher
-      gelscht. */
-  void malen(QPainter & p, int xx, int yy) const;
+      gelscht.
+      If apply_mirror is set and the level is an upside-down one, then the
+      respective coordinate-transformation is applied. */
+  void malen(int xx, int yy, bool apply_mirror=true) const;
   /** liefert die Art zurck */
   int getArt() const;
   /** liefert die Farbe zurck, aber nur, wenns wirklich ein farbiges
@@ -177,16 +195,22 @@ class Blop {
   /** Liefert zurck, wie viele Punkte dieser Stein zur Kettengre
       beitrgt (normalerweise 1). */
   int getKettenBeitrag() const;
+  /** Liefert zurck, ob ein bestimmtes Platzverhalten vorliegt */
+  bool getVerhalten(int verhalten) const;
+  /** Liefert zurck, welchem Spieler (0 oder 1) der Blop gehoert oder
+      2 bei globalblop */
+  int getSpieler() const;
 
   /** Liefert einen String der Art "Blop Drachen at x,y".
       Fr Fehlermeldungen. */
-  __String toString() const;
+  Str toString() const;
 
   /** Wird vom Code aufgerufen, wenn es Punkte geben soll. */
   void bekommPunkte(int pt) const;
   /** Wird vom Code aufgerufen, wenn ein Message angezeigt werden soll. */
-  void zeigMessage(__String mess) const;
-  
+  void zeigMessage(Str mess) const;
+
+  const ort_absolut & getOrt() const;
 
  protected:
   /* Achtung! Beim Einfgen neuer Felder muss auch aendern(...) gendert werden.
@@ -197,9 +221,6 @@ class Blop {
       gesetzt. */
   int mKettenGroesse;
   
-  /** Der aktuelle Platz-Zustand. 0 bedeutet nicht platzen. */
-  int mAmPlatzen;
-
   /***** Variablen, die sich nie ndern *****/
   /* Sie sollten am besten const sein; das geht aber aus technischen Grnden
      beim BlopGitter nicht.
@@ -212,17 +233,13 @@ class Blop {
   bool mFreilaufend;
   /** ggf. Pointer auf den Besitzer dieses Blops */
   BlopBesitzer * mBesitzer;
-  /** Welchem Spieler wir gehren, will vielleicht das Animations-Programm
-      wissen. */
-  bool mRechterSpieler;
-  /** Koordinaten des Blops im BlopGitter... oder sonstige Informationen,
-      die den BlopBesitzer interessieren. (Beim Fall ist mX 0 oder 1, je
-      nachdem, welcher der beiden Blops es ist.) */
-  int mX, mY;
+
+  ort_absolut mOrt;
   
  protected: // Protected methods
-  /** Liest die aktuelle Sorte dieses Blops aus den Leveldaten aus. */
-  Sorte * getSorte() const;
+  /** Liest die aktuelle Sorte (oder die vom Beginn der Gleichzeit) dieses
+      Blops aus den Leveldaten aus. */
+  Sorte * getSorte(bool vergangenheit = false) const;
   /** Fragt beim Besitzer an, in welche Richtungen dieser Blop
       verbunden werden kann und liefert das zurck. */
   int getVerbindungen() const;
@@ -255,7 +272,7 @@ protected:
   /** Die Zeitnummer, zu dessen Anfang mDatenAlt gehrt. */
   long mZeitNummerDatenAlt;
   /** Zeitnummer fr initStapel-Lazy-Aufruf. */
-  long mZeitNummerInitStapel;
+  long mZeitNummerLeereStapel;
   
   /** Wird nur innerhalb von animiere() gebraucht, whrend ein
       Cual-Code luft. True, wenn der Cual-Code malen darf. */
@@ -265,7 +282,6 @@ protected:
   BildStapel mBild;
   /** Der nicht mehr ganz aktuelle Bildstapel */
   BildStapel mBildAlt;
-  
 
 public:
   
@@ -274,11 +290,11 @@ public:
   void initSchritt();
 
 
-  int getSpezConst(int vnr) const;  
+  int getSpezConst(int vnr, bool vergangenheit = false) const;  
 
   /** High-Level; Wird benutzt, wenn eine Variable im cual-Programm steht.
       Kmmert sich auch um all das @()-Zeug und die Zeitverzgerung. */
-  int getVariable(const Variable & v) const;
+  int getVariable(const Variable & v);
   void setVariable(const Variable & v, int wert, int op);
   
   /** Low-Level; wird von den High-Level-Funktionen aufgerufen und vom
@@ -292,6 +308,9 @@ private:
   /** Noch low-Levler: Speichert alte Werte nicht ab, wie sich das
       innerhalb einer Gleichzeit gehren wrde. */
   void setVariableIntern(int vnr, int wert, int op);
+  /** spezvar_kind ist eine ganz spezielle spezvar: Ganz lowlevel passiert
+      hier noch mehr. */
+  void setKindIntern(int wert);
 
 
 public:
@@ -332,10 +351,10 @@ public:
   void merkeAlteVarWerte();
   
   /** Schaut, ob noch ein Lazy-Evaluation-initStapel()-Aufruf aussteht
-      und fhrt ihn ggf. aus. (Siehe lazyInitStapel().)
+      und fhrt ihn ggf. aus. (Siehe lazyLeereStapel().)
       Ist zwar eigentlich nicht wirklich const, wird aber in Situationen
       aufgerufen, wo man sich eigentlich const fhlen mchte. */
-  void braucheInitStapel() const;
+  void braucheLeereStapel() const;
   
   /** Speichert das aktuelle Bild (d. h. aus den spezvar file und pos)
       in die Mal-Liste */
@@ -343,41 +362,32 @@ public:
 
   /** Speichert das aktuelle Bild (d. h. aus den spezvar file und pos)
       in die Mal-Liste von einem anderen Blop, und zwar so, dass es
-      erst zum Schluss gemalt wird. */
-  void speichereBildRelativ(int xx, int yy);
-  
-  
-  /********** Init-Event-Verwaltung **********/
-
-  /* Funktionsweise:
-     Blops, die neu erzeugt werden, tragen sich mit setWillInitEvent()
-     in die Liste derer ein, die gerne noch ein Init-Event htten.
-     In Cuyo::animiere() wird dann vor der Animation fr alle eingetragenen
-     Blops in einer Gleichzeit das Init-Event aufgerufen.
-     
-     Nachteil: Ein frisch erzeugter Blop sollte noch keine anderen Events
-     bekommen, solange noch nicht einmal die Hauptschleife passiert wurde.
-   */
+      in Ebene ebene gemalt wird. */
+  void speichereBildFremd(Ort & ort, int ebene);
 
 protected:
 
-  /** True, wenn dieser Blop noch kein Init-Event bekommen hat. Nur
-      mit der nachfolgenden Methode setzen und lschen, damit das auch
-      global gespeichert wird. */
-  bool mWillInitEvent;
+  /** Nummer eines Events, das der Blop noch erwartet; event_keins, wenn
+   *  der Blop kein Event erwartet. Wert nur mit den Methoden
+   *  scheduleEvent() und unscheduleEvent() ndern, damit das auch
+   *  in der globalen Liste aktualisiert wird.
+   */
+  int mScheduleEventNr;
 
-  /** Wo in gWIEListe steht dieser Blop (falls er ein Init-Event will) */
-  int mWIEPos;
+  /** Wo in gSEListe steht dieser Blop (falls er ein Event will) */
+  int mSEPos;
 
-  /** Setzt, ob dieser Blop ein Init-Event will. */
-  void setWillInitEvent(bool wie);
+  /** Lscht diesen Blop aus gWEListe (falls ntig) */
+  void unscheduleEvent();
 
-  /** Liste der Blops, die ein Init-Event wollen. */
-  static QArray<Blop *> gWIEListe;
+  /** Liste der Blops, die ein Event wollen. */
+  static std::vector<Blop *> gSEListe;
 
 public:
-  /** Ruft alle noch fehlenden Init-Events aus. */
-  static void sendeFehlendeInitEvents();
+  /** Ruft alle geschedulten Events aus. */
+  static void sendeGeschedulteEvents();
+
+  void playSample(int nr) const;
 
 
   /********** Statisches Zeug **********/
@@ -405,7 +415,7 @@ public:
       d. h. erst dann, wenn's wirklich gebraucht wird.
       Gebraucht wird's natrlich, wenn ein Blop animiert wird. Aber auch,
       wenn ein Nachbarblop etwas auf diesen Blop malt. */
-  static void lazyInitStapel();
+  static void lazyLeereStapel();
   
   /** Stellt ein, ob Blops */
   //static void setInitEventsAutomatisch(bool iea);
@@ -421,10 +431,10 @@ protected:
       passieren Fehler... man knnte diese Variable vor jedem Level frisch
       initialisieren.) */
   static long gAktuelleZeitNummerDatenAlt;
-  /** Entsprechend fr initStapel()-Aufrufe. D. h.: Wenn mZeitNummerInitStapel
-      eines Blops kleiner als gAktuelleZeitNummerInitStapel ist, muss
+  /** Entsprechend fr initStapel()-Aufrufe. D. h.: Wenn mZeitNummerLeereStapel
+      eines Blops kleiner als gAktuelleZeitNummerLeereStapel ist, muss
       initStapel() noch aufgerufen werden. */
-  static long gAktuelleZeitNummerInitStapel;
+  static long gAktuelleZeitNummerLeereStapel;
   
   /** True heit, dass neu erzeugte Blops automatish init-Events
       bekommen. (true ist default) */
@@ -445,7 +455,7 @@ protected:
 
   /** Liste der Zukunfts-Zuweisungen, die sich whrend einer Gleichzeit
       ansammeln. */
-  static QArray<tZZ> gZZ;
+  static std::vector<tZZ> gZZ;
   static int gZZAnz;
 
 
