@@ -4,6 +4,9 @@
     begin                : Sun Jul 1 2001
     copyright            : (C) 2001 by Immi
     email                : cuyo@karimmi.de
+
+Modified 2002,2005,2006,2010,2011,2014 by the cuyo developers
+
  ***************************************************************************/
 
 /***************************************************************************
@@ -22,12 +25,11 @@
 
 
 /** In parser.yy definiert. */
-void parse(__String name, DefKnoten * erg);
+void parse(const Str & name, DefKnoten * erg);
 
 
 DatenDatei::DatenDatei():
-  mDaten(new DefKnoten()), /* Leerer Knoten ohne Vater */
-  mEinSpielerVersion(false)
+  mDaten(new DefKnoten()) /* Leerer Knoten ohne Vater */
 {
   /* Das Squirrel sitzt am Anfang ganz oben */
   initSquirrel();
@@ -54,17 +56,11 @@ void DatenDatei::leeren() {
 
 /** Ldt die angegebene Datei. (Kann mehrmals aufgerufen werden, um
     mehrere Dateien gleichzeitig zu laden.) */
-void DatenDatei::laden(__String name) {
+void DatenDatei::laden(const Str & name) {
   parse(name, mDaten);
 }
 
 
-/** Setzt, ob nach Ein-Spieler-Version-Schlsseln
-gesucht werden soll */
-void DatenDatei::setEinSpielerVersion(bool esv /*= false*/) {
-  mEinSpielerVersion = esv;
-}
-
 
 
 
@@ -88,14 +84,14 @@ bool DatenDatei::existiertSquirrelKnoten() const {
 
 
 /** Liefert die Position des Squirrels als String. */
-__String DatenDatei::getSquirrelPosString() const {
+Str DatenDatei::getSquirrelPosString() const {
   return mSquirrelPosString;
 }
 
 
 /** Das Eichhrnchen klettert weiter weg von der Wurzel. Wird von
     DatenDateiPush benutzt. */
-void DatenDatei::kletterWeiter(__String na) {
+void DatenDatei::kletterWeiter(const Str & na, const Version & version) {
 
   /* Neuen Abschnittnamen bauen */
   if (!mSquirrelPosString.isEmpty())
@@ -103,7 +99,8 @@ void DatenDatei::kletterWeiter(__String na) {
   mSquirrelPosString += na;
 
   /* Knoten zum neuen Abschnitt suchen */
-  mSquirrelKnoten = (DefKnoten *) getEintragKnoten(na, type_DefKnoten);
+  mSquirrelKnoten = (DefKnoten *) getEintragKnoten(na, version, true,
+						   type_DefKnoten);
   
   /* Wenn dieser Unterabschnitt existiert, dann auch die Codeen
      dort suchen. */
@@ -127,28 +124,26 @@ DefKnoten * DatenDatei::getSquirrelPos() const {
 
 
 /** Liefert den angegebenen Eintrag beim Squirrel.
-    Macht ggf. 1-Spieler-Modus-Schlsselumbenennungen.
     Prft, ob der Typ der gewnschte ist.
-    Liefert 0, wenn's den Eintrag nicht gibt.
+    Liefert 0, wenn's den Eintrag nicht gibt, aber defaultVorhanden.
     Throwt bei sonstigem Fehler. */
-Knoten * DatenDatei::getEintragKnoten(__String schluessel, int typ) const {
+Knoten * DatenDatei::getEintragKnoten(const Str & schluessel,
+				      const Version & version,
+				      bool defaultVorhanden, int typ) const {
 
-  if (!mSquirrelKnoten)
-    return 0;
+  if (!mSquirrelKnoten) {
+    if (defaultVorhanden)
+      return 0;
+    else throw Fehler("%s required but not defined",schluessel.data());
+  }
+  
+  Knoten * ret = mSquirrelKnoten->getKind(schluessel,version,
+					  defaultVorhanden);
 
-  /* Im Einspielermodus schauen, ob das Schlsselwort mit 1 hintendran
-     existiert. Wenn ja, diese Version nehmen */
-  if (mEinSpielerVersion && mSquirrelKnoten->enthaelt(schluessel+'1'))
-    schluessel += '1';
-  
-  if (!mSquirrelKnoten->enthaelt(schluessel))
-    return 0;
-  
-  Knoten * ret = mSquirrelKnoten->getKind(schluessel);
-  
-  if (typ != type_egal && ret->type() != typ)
-    throw Fehler(_("Wrong type on the righthand side of %s="),
-                 schluessel.data());
+  if (ret)
+    if (typ != type_egal && ret->type() != typ)
+      throw Fehler("Wrong type on the righthand side of %s%s=",
+                   schluessel.data(), version.toString().data());
     
   return ret;
 }
@@ -157,68 +152,118 @@ Knoten * DatenDatei::getEintragKnoten(__String schluessel, int typ) const {
 	
 
 /** Gibt's den Eintrag? */
-bool DatenDatei::hatEintrag(__String schluessel) const {
-  
-  return getEintragKnoten(schluessel, type_egal) != 0;
+bool DatenDatei::hatEintrag(const Str & schluessel) const {
+  return mSquirrelKnoten->enthaelt(schluessel);
 }
 
-/** Liefert den Eintrag, wenn er existiert, sonst null.
-    Schaut ggf. nach EinSpieler-Eintrag. */
-const DatenKnoten * DatenDatei::getEintrag(__String schluessel,
+/** Liefert den Eintrag, wenn er existiert, sonst null, wenn
+    defaultVorhanden, sonst wird gethrowt. */
+const DatenKnoten * DatenDatei::getEintrag(const Str & schluessel,
+					   const Version & version,
+					   bool defaultVorhanden,
 					   int typ
 					   /*= type_EgalDatum*/) const {
-  Knoten * e = getEintragKnoten(schluessel, type_ListenKnoten);
+  Knoten * e = getEintragKnoten(schluessel, version, defaultVorhanden,
+				type_ListenKnoten);
   if (e)
     return ((ListenKnoten *) e)->getDatum(0)->assert_datatype(typ);
   else
     return 0;
 }
 
-/** Dito fr Wrter, default fr default ist null. */
-__String DatenDatei::getWortEintrag(__String schluessel,
-				    __String def /*= __String()*/) const {
-  const DatenKnoten * e = getEintrag(schluessel,type_WortDatum);
+/** Dito fr Wrter. Default fr def ist "". */
+Str DatenDatei::getWortEintragOhneDefault(const Str & schluessel,
+					       const Version & version) const {
+  return getEintrag(schluessel,version,false,type_WortDatum)->getWort();
+}
+
+Str DatenDatei::getWortEintragMitDefault(const Str & schluessel,
+					      const Version & version,
+					      Str def /*= Str()*/) const {
+  const DatenKnoten * e = getEintrag(schluessel,version,true,type_WortDatum);
   if (e)
     return e->getWort();
   else
     return def;
 }
 
-/** Liefert den Eintrag als Zahl, wenn er existiert, sonst die default-Zahl. */
-int DatenDatei::getZahlEintrag(__String schluessel, int def /*= 0*/) const {
-  const DatenKnoten * e = getEintrag(schluessel,type_ZahlDatum);
+/** Liefert den Eintrag als Zahl. */
+int DatenDatei::getZahlEintragOhneDefault(const Str & schluessel,
+					  const Version & version) const {
+  return getEintrag(schluessel,version,false,type_ZahlDatum)->getZahl(0);
+}
+
+int DatenDatei::getZahlEintragMitDefault(const Str & schluessel,
+					 const Version & version,
+					 int def /*= 0*/) const {
+  const DatenKnoten * e = getEintrag(schluessel,version,true,type_ZahlDatum);
   if (e) {
     return e->getZahl(0);
   } else
     return def;
 }
 
-/** Liefert den Eintrag als Farbe, wenn er existiert,
-    sonst die default-Farbe. */
-QColor DatenDatei::getFarbEintrag(__String schluessel,
-				   const QColor & def /*= __black*/) const {
-  ListenKnoten * e = getListenEintrag(schluessel);
-  
+bool intZuBool(int i) {
+  switch (i) {
+  case 0: return false;
+  case 1: return true;
+  default: throw Fehler("0 or 1 expected, got %d.",i);
+  }
+}
+
+bool DatenDatei::getBoolEintragOhneDefault(const Str & schluessel,
+					   const Version & version) const {
+  return intZuBool(getZahlEintragOhneDefault(schluessel,version));
+}
+
+
+bool DatenDatei::getBoolEintragMitDefault(const Str & schluessel,
+					  const Version & version,
+					  bool def) const {
+  return intZuBool(getZahlEintragMitDefault(schluessel,version,(def ? 1 : 0)));
+}
+
+
+/** Liefert den Eintrag als Farbe. */
+Color DatenDatei::getFarbEintragOhneDefault(const Str & schluessel,
+					    const Version & version) const {
+  ListenKnoten * e = getListenEintrag(schluessel,version,false);
+  if (e->getLaenge() != 3)
+    throw Fehler("%s","Color (r,g,b) expected");
+            
+  return Color(e->getDatum(0)->assert_datatype(type_ZahlDatum)->getZahl(),
+		e->getDatum(1)->assert_datatype(type_ZahlDatum)->getZahl(),
+		e->getDatum(2)->assert_datatype(type_ZahlDatum)->getZahl());
+}
+
+Color DatenDatei::getFarbEintragMitDefault(const Str & schluessel,
+				  const Version & version,
+				  const Color & def /*= black*/) const {
+  ListenKnoten * e = getListenEintrag(schluessel,version,true);
+
   if (e == 0) return def;
   
   if (e->getLaenge() != 3)
-    throw Fehler(_("Color (r,g,b) expected"));
+    throw Fehler("%s","Color (r,g,b) expected");
             
-  return QColor(e->getDatum(0)->assert_datatype(type_ZahlDatum)->getZahl(),
+  return Color(e->getDatum(0)->assert_datatype(type_ZahlDatum)->getZahl(),
 		e->getDatum(1)->assert_datatype(type_ZahlDatum)->getZahl(),
 		e->getDatum(2)->assert_datatype(type_ZahlDatum)->getZahl());
 }
 
 /** Liefert einen Eintrag als Knoten */
-ListenKnoten * DatenDatei::getListenEintrag(__String schluessel) const {
-  Knoten * e = getEintragKnoten(schluessel, type_ListenKnoten);
+ListenKnoten * DatenDatei::getListenEintrag(const Str & schluessel,
+					    const Version & version,
+					    bool defaultVorhanden) const {
+  Knoten * e = getEintragKnoten(schluessel, version, defaultVorhanden,
+				type_ListenKnoten);
   if (e) {
     /* Frher freute sich der Aufrufer darber, da er wusste, da die
        Liste nur Wrter enthlt. Es knnten noch Probleme existieren */
     ListenKnoten * ret = (ListenKnoten*) e;
     for (int i = 0; i < ret->getLaenge(); i++)
       if (ret->getKind(i)->type() != type_DatenKnoten)
-        throw Fehler(_("List of atomic data expected"));
+        throw Fehler("%s","List of atomic data expected");
     return ret;
   } else
     return 0;
@@ -227,15 +272,30 @@ ListenKnoten * DatenDatei::getListenEintrag(__String schluessel) const {
 
 
 /** Sucht einen Code beim Squirrel oder nher an der Wurzel.
-    Behlt den Besitz am Code.
+    Behlt den Besitz am Code/an der VarDefinition.
     Throwt bei nicht-existenz. */
-Code * DatenDatei::getCode(int ns, __String name) {
+Code * DatenDatei::getCode(const Str & name,
+			   const Version & version, bool defaultVorhanden) {
 
   /* Einen Codeabschnitt sollte es eigentlich immer geben - zumindest
      wenn die Datei auf ist. Aber die sollte immer auf sein. */
   CASSERT(mSquirrelCodeKnoten);
 
-  return (Code *) mSquirrelCodeKnoten->getDefinition(ns, name);
+  return (Code *) mSquirrelCodeKnoten->getDefinition(namespace_prozedur,
+						     name, version,
+						     defaultVorhanden);
+}
+
+VarDefinition * DatenDatei::getVarDef(const Str & name,
+				      const Version & version,
+				      bool defaultVorhanden) {
+
+  /* Einen Codeabschnitt sollte es eigentlich immer geben - zumindest
+     wenn die Datei auf ist. Aber die sollte immer auf sein. */
+  CASSERT(mSquirrelCodeKnoten);
+
+  return (VarDefinition *) mSquirrelCodeKnoten->getDefinition(
+      namespace_variable, name, version, defaultVorhanden);
 }
 
 
@@ -248,14 +308,15 @@ Code * DatenDatei::getCode(int ns, __String name) {
 
 
 
-DatenDateiPush::DatenDateiPush(DatenDatei & c, __String name,
+DatenDateiPush::DatenDateiPush(DatenDatei & c,
+			       const Str & name, const Version & version,
                                bool verlange /*= true*/): mConf(c) {
   mMerkName = mConf.mSquirrelPosString;
   mMerkKnoten = mConf.mSquirrelKnoten;
   mMerkCodeKnoten = mConf.mSquirrelCodeKnoten;
-  mConf.kletterWeiter(name);
+  mConf.kletterWeiter(name,version);
   if (verlange && !mConf.existiertSquirrelKnoten())
-    throw Fehler(_("Section %s does not exist."),
+    throw Fehler("Section %s does not exist.",
                  mConf.getSquirrelPosString().data());
 }
 
