@@ -4,6 +4,9 @@
     begin                : Sat Aug 18 2001
     copyright            : (C) 2001 by Immi
     email                : cuyo@karimmi.de
+
+Modified 2001,2002,2005,2006,2008,2010,2011,2014 by the cuyo developers
+
  ***************************************************************************/
 
 /***************************************************************************
@@ -18,10 +21,12 @@
 #ifndef FALL_H
 #define FALL_H
 
-#include "inkompatibel.h"
+#include <SDL.h>
 
 #include "blop.h"
-
+#include "blopbesitzer.h"
+#include "bilddatei.h"
+#include "leveldaten.h"
 
 
 // wie das Fallende sein kann
@@ -29,6 +34,7 @@
 #define richtung_einzel 1
 #define richtung_waag 2
 #define richtung_senk 3
+#define richtung_unplatziert 4 // das Fall, das als naechstes kommen wird...
 
 // Wert fr get_X bis get_YY, wenn das Blop nicht existiert
 #define blop_pos_nix -1
@@ -36,37 +42,23 @@
 
 
 struct FallPos {
-  /** x-Koord vom Fall in Feldern */
+  /** x-Koord vom Fall in Feldern; bei unplatzierten Blops ist das
+      die Pos, an der der Blop spter auftauchen wird. */
   int x;
   /** y-Koord vom Fall in Pixeln (absolut, und nicht relativ zum ggf.
-      wegen Rberreihen verschobenen Spielfeld) */
+      wegen Rberreihen verschobenen Spielfeld).
+      Bei unplatzierten blobs ist das die Pos. relativ zum Preview-Feld 
+      (und nicht die Pos., an der es auftauchen wird) */
   int yy;
   /** Richtung und Anzahl der Blops. Siehe Konstanten. */
   int r;
 	
   /** Liefert die Anzahl der Blops vom Fall zurck. */
-  int getAnz() const {
-    if (r == richtung_keins)
-      return 0;
-    else if (r == richtung_einzel)
-      return 1;
-    else
-      return 2;
-  }
+  int getAnz() const;
 	
-  int getX(int a) const {
-    return x + a * (r == richtung_waag);
-  }
+  int getX(int a) const;
 
-  /** ... hv = aktuelle Hochverschiebung... */
-  int getY(int a, int hv) const {
-    int y0 = yy + gric - 1 + hv;
-    /* Korrektur bei ungeraden Spalten im Sechseck-Modus... */
-    if (ld->mSechseck && (getX(a) & 1))
-      y0 += gric / 2;
-			
-    return y0 / gric + a * (r == richtung_senk);
-  }
+  int getY(int a, const Spielfeld * spf) const;
 
 };
 
@@ -75,7 +67,6 @@ class Spielfeld;
 
 /** Enthlt alle Informationen ber das Fall: Existenz, Position
 		(inkl. genaue Drehpos), Farben, ...
-  *@author Immi
   */
 
 class Fall: public BlopBesitzer {
@@ -95,10 +86,15 @@ class Fall: public BlopBesitzer {
   bool mSchnell;
   /** Ist es noch nicht ganz fertiggedreht? */
   int mExtraDreh;
-  /** Ist es noch nicht ganz fertig waagrecht verschoben? */
+  /** Ist es noch nicht ganz fertig waagerecht verschoben? */
   int mExtraX;
+  /** Wurde Taste links bzw. rechts gedrckt? */
+  int mExtraLinks, mExtraRechts;
   /** Die Blops */
   Blop mBlop[2];
+  /** Rechteck, das den Bereich ueberdeckt, wo das Fall im Moment
+      hingemalt ist (fuer Grafik-Updates). */
+  int mRectX, mRectY, mRectW, mRectH;
 
 	
   /** kopiert einen fallenden Blop nach mDaten und liefert den
@@ -115,19 +111,11 @@ class Fall: public BlopBesitzer {
     return mPos.r == richtung_waag || mPos.r == richtung_senk;
   }
 	
-  /** Setzt die Grafik vom Fall auf upzudaten */
-  void setUpdate();
-
   /** Prueft, ob an Position p schon was im Weg ist oder nicht.
-      Wenn irgendwo drber in der Spalte was ist, zhlt das auch
-      als im Weg.
+      Wenn irgendwo drber in der Spalte was nicht-schwebendes ist,
+      zhlt das auch als im Weg.
       @return eine Konstante belegt_... */
   int testBelegt(FallPos p) const;
-
-  /** Liefert die Anzahl der Bklops vom Fall zurck. */
-  int getAnz() const {
-    return mPos.getAnz();
-  }
 	
   /** Liefert true, wenn das Fallende senkrecht ist */
   bool istSenkrecht() const;
@@ -135,80 +123,94 @@ class Fall: public BlopBesitzer {
   /** Lsst nur noch Blop a brig */
   void halbiere(int a);
 	
+  /** Liefert loc_x */
   int getX(int a) const;
+  /** Liefert loc_y */
   int getY(int a) const;
 
+  /** Liefert loc_xx */
   int getXX(int a) const;
+  /** Liefert loc_yy */
   int getYY(int a) const;
 
   /** Kodiert alle Informationen des Drehens in eine Zahl.
       Ist fr getXX und getYY da */
   int getDrehIndex(int a) const;
 
- public:
+  /** Bestimm mFallRect neu (Fall-ueberdeckendes Rechteck) */
+  void calcFallRect();
 
+  /** Setzt den Bereich, der durch mFallRect angegeben ist, auf
+      upzudaten */
+  void setUpdateFallRect();
+
+ public:
+ 
   /**  Mu einmal aufgerufen werden */
   void initialisiere();
 
-  /** Erzeugt ein neues Fall. Liefert false, wenn dafr kein Platz ist */
-  bool erzeug();
+  /** Erzeugt ein neues Fall, der sich noch nicht im Spielfeld befindet.
+      Bei reinkommen = true kommt das Fall ins Preview-Feld reingerutscht.
+      Sonst ist es sofort da (fr Spielanfang) */
+  void erzeug(bool reinkommen = true);
+  /** Bringt ein Fall ins Spiel, d.h. setzt die Koordinaten. Liefert false, wenn dafr kein Platz ist */
+  bool insSpiel();
   /** Entfernt das Fall ganz */
   void zerstoere();
 
 	
-  /** Liefert ein Rechteck zurck, dass das Fallende berdeckt */
-  QRect getRect() const {
-    /* Faule Variante... ohne Beachtung der Drehposition und der genauen
-       y-Position */
-    int li = mPos.x;
-    int re = mPos.x + 2;
-    if (mExtraX < 0) li--;
-    if (mExtraX > 0) re++;
-    return QRect(li * gric, mPos.yy, (re - li) * gric, 3 * gric);
-    /*return QRect(x * gric, y * gric,
-      (1 + (r == richtung_waag)) * gric,
-      (1 + (r == richtung_senk)) * gric);
-    */
-  }
-	
+ private:
+  /** Fuer platzierte Falls: Macht alles von spielSchritt ausser Grafik-Update markieren */
+  void spielSchrittPlatziertIntern();
+  
+ public:
   /** Bewegt das Fall ggf. nach unten und kmmert sich ggf. um
       Verwandlungen. */
   void spielSchritt();
   
   /** Fhrt die Animationen durch. Innerhalb einer Gleichzeit aufrufen. */
-  void  animiere();
+  void animiere();
 
-  /** Bewegt das Fall eins nach links */
+  /** Bewegt das Fall eins nach links. Wird in einer Gleichzeit aufgerufen. */
   void tasteLinks();
-  /** Bewegt das Fall eins nach rechts */
+  /** Bewegt das Fall eins nach rechts. Wird in einer Gleichzeit aufgerufen. */
   void tasteRechts();
-  /** Dreht das Fall */
-  void tasteDreh();
-  /** ndert die Fallgeschwindigkeit vom Fall */
+  /** Dreht das Fall.
+      Version 1 wird in einer Gleichzeit aufgerufen,
+      Version 2 danach auerhalb von Gleichzeiten. */
+  void tasteDreh1();
+  void tasteDreh2();
+  /** ndert die Fallgeschwindigkeit vom Fall.
+      Wird in einer Gleichzeit aufgerufen. */
   void tasteFall();
 
   /** Liefert true, wenn das Fall (noch) am Platzen ist
       (wg. Spielende) */
   bool getAmPlatzen() const;
   /** Liefert einen Pointer auf die Blops zurck. Wird vom
-      KIPlayer und von @0 und @1 bentigt. */
+      KIPlayer und von ort_absolut::finde() bentigt. */
   const Blop * getBlop() const;
   Blop * getBlop();
+  /** Liefert die Anzahl der Blops vom Fall zurck. */
+  int getAnz() const {
+    return mPos.getAnz();
+  }
   /** Lsst alle Blops vom Fall platzen (Spielende). */
   void lassPlatzen();
   /** Malt das Fall. */
-  void malen(QPainter & p) const;
+  void malen() const;
   /** Liefert true, wenn das Fall existiert.
       Mit Argument: Wenn diese Hlfte existiert. */
   bool existiert(int a=0) const;
-  /** Liefert true, wenn das Fall aus grade am zerfallen ist
+  /** Liefert true, wenn das Fall gerade am zerfallen ist
       (d. h. existiert, aber aus nur noch einem Blop besteht).
       In dieser Zeit darf nmlich keine Explosion gezndet
       werden. (Erst warten, bis der andere Blop auch angekommen
       ist.) */
-  bool istEinzel();
+  bool istEinzel() const;
+  void playSample(int nr) const;
 
-  virtual int getSpezConst(int vnr) const;
+  virtual int getSpezConst(int vnr, const Blop *) const;
 
 };
 
