@@ -4,6 +4,9 @@
     begin                : Wed Jul 25 2001
     copyright            : (C) 2001 by Immi
     email                : cuyo@karimmi.de
+
+Modified 2001-2003,2006,2011 by the cuyo developers
+
  ***************************************************************************/
 
 /***************************************************************************
@@ -18,8 +21,6 @@
 #ifndef KIPLAYER_H
 #define KIPLAYER_H
 
-#include "inkompatibel.h"
-
 /**Wenn man gegen den Computer spielt...
   *@author Immi
   */
@@ -35,16 +36,14 @@ public:
   /** Teilt der KI mit, dass ein neuer Level anfngt. (Initialisiert
       alles.) */
   void startLevel();
-public slots: // Public slots
+public:
   /** Einmal pro Spielschritt aufrufen, wenn der Computer
-auch spielen soll */
+      auch spielen soll */
   void spielSchritt();
 
 protected:
   Spielfeld * mSp;
   const BlopGitter * mDaten;
-  /** Zeigt auf ein Array von zwei Blops... */
-  const Blop * mFall;
 
   bool mZuTun;
   double mNochWart;
