@@ -5,9 +5,9 @@
     copyright            : (C) 2000 by Immi
     email                : cuyo@pcpool.mathematik.uni-freiburg.de
 
-Modified 2001-2006,2008-2011 by the cuyo developers
+Modified 2001-2006,2008-2011,2014 by the cuyo developers
 Modified 2012 by Bernhard R. Link
-Maintenance modifications 2012 by the cuyo developers
+Maintenance modifications 2012,2014 by the cuyo developers
 
  ***************************************************************************/
 
@@ -46,6 +46,57 @@ Maintenance modifications 2012 by the cuyo developers
 using namespace std;
 
 
+
+/***************************************************************************/
+
+
+SortenArray::SortenArray(): mLaenge(-1) {
+}
+
+
+/** Sorte auslesen */
+Sorte * SortenArray::operator[](int nr) const {
+  return mSorten[nr - blopart_min_sorte];
+}
+
+
+/** Länge des Arrays ändern (genauer gesagt: Länge des positiven Teils);
+ *  sollte nur aufgerufen werden, wenn das Array leer ist. */
+void SortenArray::setLaenge(int l) {
+  CASSERT(mLaenge == -1);
+  CASSERT(l <= max_farben_zahl);
+  mLaenge = l;
+  for (int i = 0; i < mLaenge - blopart_min_sorte; i++)
+    mSorten[i] = NULL;
+}
+
+
+/** Sorte in das Array einfügen; normalerweise wird die Sorte dann
+    von loeschen() gelöscht; aber nicht, wenn istOriginal = false ist;
+    dann wird davon ausgegangen, dass der Pointer nur eine Kopie ist. */
+void SortenArray::neueSorte(int nr, Sorte * s, bool istOriginal /*= true*/)  {
+  CASSERT(mLaenge != -1 && nr < mLaenge);
+  nr -= blopart_min_sorte;
+  CASSERT(mSorten[nr] == NULL);
+  mSorten[nr] = s;
+  mOriginale[nr] = istOriginal;
+}
+  
+  
+void SortenArray::loeschen() {
+  if (mLaenge == -1) return;
+  for (int i = 0; i < mLaenge - blopart_min_sorte; i++) {
+    if (mOriginale[i]) delete mSorten[i];
+    mSorten[i] = NULL;
+  }
+  mLaenge = -1;
+}
+
+
+
+/***************************************************************************/
+
+
 /* Globale Variable mit den Level-Daten */
 
 LevelDaten * ld;
@@ -54,8 +105,6 @@ LevelDaten * ld;
 LevelDaten::LevelDaten(const Version & version): mLevelCache(),
   mVersion(version),
   mLevelGeladen(false),
-/* Wir basteln uns ein Array, dessen Indizierung nicht mit 0 beginnt...: */
-  mSorten(mSortenIntern - blopart_min_sorte),
   /* Noch keine Sorten geladen */
   mAnzFarben(0)
 /* Uninitialisierte Blops entstehen sehr frueh, und die haben auch schon
@@ -68,10 +117,6 @@ LevelDaten::LevelDaten(const Version & version): mLevelCache(),
   }
   ld = this;
   
-  /* Auch die negativen Sorten sind noch nicht geladen: */
-  for (int i = blopart_min_sorte; i < 0; i++)
-    mSorten[i] = NULL;
-
   /* So; jetzt haben die Variablen einen gueltigen Zustand,
      und wir koennen schon mal erste Dinge laden. */
   ladLevelSummary(true,version);
@@ -253,19 +298,17 @@ void LevelDaten::ladSorten(const Str & ldKeyWort, int blopart) {
     int nr = mLevelConf[ldteil_level]->getSquirrelPos()->
       getSortenAnfang(ldKeyWort);
     if (nr+neueFarben > max_farben_zahl)
-      /* TRANSLATORS: "pics" is a programming keyword and should not be
-	 translated */
-      throw Fehler(_sprintf(_("Number of pics exceeds limit %d"),
+      throw Fehler(_sprintf("Number of pics exceeds limit %d",
 			    max_farben_zahl));
 
     for (int namen_nr = 0; namen_nr < neueNamen; namen_nr++) {
-      mSorten[nr] = new Sorte(picsnamen->getKernDatum(namen_nr,type_WortDatum)
+      mSorten.neueSorte(nr, new Sorte(picsnamen->getKernDatum(namen_nr,type_WortDatum)
 		                ->getWort(),
-			      mVersion, blopart);
+			      mVersion, blopart));
       nr++;
  
       for (int i = picsnamen->getVielfachheit(namen_nr)-1; i>0; i--) {
-	mSorten[nr]=mSorten[nr-1];
+	mSorten.neueSorte(nr, mSorten[nr-1], false); // false = ist nur kopie
 	nr++;
       }
     }
@@ -283,20 +326,10 @@ void LevelDaten::entladLevel() {
      halbgeladenen Zustand; dann wuerden wir auch gerne allen
      Speicher freigeben, der grad belegt war */
   
-  /* Positive Sorten loeschen */
-  for (int bnr=mAnzFarben-1; bnr>0; bnr--)
-    if (bnr == 0 || mSorten[bnr]!=mSorten[bnr-1]) {
-      if (mSorten[bnr])  // Koennte NULL sein falls ladefehler
-        delete mSorten[bnr];
-    }
-  mAnzFarben = 0;
+  /* Alle Sorten löschen */
+  mSorten.loeschen();
 
-  /* Die negativen Sorten loeschen */
-  for (int bnr = blopart_min_sorte; bnr < 0; bnr++)
-    if (mSorten[bnr]) {
-      delete mSorten[bnr];
-      mSorten[bnr] = NULL;
-    }
+  mAnzFarben = 0;
 }
 
 
@@ -305,7 +338,7 @@ void LevelDaten::entladLevel() {
 void LevelDaten::ladLevel(int nr) {
 
   if (!mLCGeladen[ldteil_summary])
-    throw Fehler("%s",_("Sorry, no working level description file available."));
+    throw Fehler("%s","Sorry, no working level description file available.");
 
   /* Ggf. Speicher von altem Level freigeben */
   entladLevel();
@@ -325,7 +358,7 @@ void LevelDaten::ladLevel(int nr) {
   ladLevelConfig();
 
   if (!mLCGeladen[ldteil_level])
-    throw Fehler("%s",_("Sorry, no working level description file available."));
+    throw Fehler("%s","Sorry, no working level description file available.");
 
   try {
 
@@ -335,8 +368,7 @@ void LevelDaten::ladLevel(int nr) {
 
     DatenDatei * daten = mLevelConf[ldteil_level];
 
-    // TRANSLATORS: This prepends an error message with position information
-    fehlerpos = _sprintf(_("At position %s (or somewhere below): "),
+    fehlerpos = _sprintf("At position %s (or somewhere below): ",
 			 daten->getSquirrelPos()->getDefString().data());
 
     /* DefKnoten des Levels abspeichern. */
@@ -387,9 +419,7 @@ void LevelDaten::ladLevel(int nr) {
     hetzrandZeit = daten->
       getZahlEintragMitDefault("toptime", mVersion, toptime_default);
     if (hetzrandZeit < 1)
-      /* TRANSLATORS: "toptime" is a programming keyword which should not
-	 be translated. */
-      throw Fehler("%s",_("toptime must be positive"));
+      throw Fehler("%s","toptime must be positive");
 
     /* Hetzrandbildchen (optional) */
     mMitHetzbildchen = daten->hatEintrag("toppic");
@@ -419,20 +449,22 @@ void LevelDaten::ladLevel(int nr) {
     mNachbarschaft = daten->
       getZahlEintragMitDefault("neighbours", mVersion, nachbarschaft_normal);
     if (mNachbarschaft < 0 || mNachbarschaft > nachbarschaft_letzte)
-      /* TRANSLATORS: "neighbours" is a programming keyword which
-	 should not be translated. */
-      throw Fehler("%s",_("neighbours out of range"));
+      throw Fehler("%s","neighbours out of range");
     /* Sechseck-Raster? */
     mSechseck =
       mNachbarschaft == nachbarschaft_6 ||
       mNachbarschaft == nachbarschaft_6_schraeg ||
       mNachbarschaft == nachbarschaft_6_3d;
-
+      
+    mSechseckFlip = daten->
+      getZahlEintragMitDefault("hexflip", mVersion, 0);
+    if (mSechseckFlip < 0 || mSechseckFlip > 3)
+      throw Fehler("%s", "hexflip out of range");
 
     /* Zufaellige Graue? (optional) */
     mZufallsGraue = daten->
       getZahlEintragMitDefault("randomgreys", mVersion, zufallsgraue_keine);
-  	  	
+
     /* Wo sind welche Grasbildchen am Anfang? */
     mAnfangsZeilen = daten->getListenEintrag("startdist",mVersion,false);
 
@@ -463,8 +495,7 @@ void LevelDaten::ladLevel(int nr) {
        gerufen wird, sondern in der Reihenfolge, wie's in der
        ld-Datei steht. */
     mAnzFarben = daten->getSquirrelPos()->getSortenAnzahl();
-    for (int i = 0; i < mAnzFarben; i++)
-      mSorten[i] = NULL;
+    mSorten.setLaenge(mAnzFarben);
 
     /* Wie viele Bilder malt ein Blop in einem Schritt hoechstens?
        Erst mal keine. Die Sorten erhoehen diese Variablen selbst, wenn
@@ -475,20 +506,35 @@ void LevelDaten::ladLevel(int nr) {
     ladSorten("pics",blopart_farbe);
     ladSorten("startpic",blopart_gras);
     ladSorten("greypic",blopart_grau);
-
+    
     /* Leer-Bildchen (optional) */
     mMitLeerBildchen = daten->hatEintrag("emptypic");
     /* Auch, wenn es kein Leer-Bildchen gibt, soll es geladen werden;
        dann liefert getEintrag() "" zurueck, und Sorte::laden() weiss,
        dass es nur alle Werte auf Defaults setzen soll. */
-    mSorten[blopart_keins] = new Sorte(daten->
+    mSorten.neueSorte(blopart_keins, new Sorte(daten->
         getWortEintragMitDefault("emptypic", mVersion, ""),
-      mVersion, blopart_keins);
+      mVersion, blopart_keins));
 
     /* Globaler Code (optional) */
-    mSorten[blopart_global] = new Sorte("global", mVersion, blopart_global);
-    mSorten[blopart_semiglobal] = new Sorte("semiglobal", mVersion,
-					    blopart_semiglobal);
+    mSorten.neueSorte(blopart_global, new Sorte("global", mVersion, blopart_global));
+    mSorten.neueSorte(blopart_semiglobal, new Sorte("semiglobal", mVersion,
+					    blopart_semiglobal));
+    mSorten.neueSorte(blopart_info, new Sorte("info", mVersion,
+					    blopart_info));
+    
+//     /* Info-Blop (optional) */
+//     if (daten->hatEintrag("infopic")) {
+//       /* to do....*/
+//     } else {
+//       int gras_nr = mLevelConf[ldteil_level]->getSquirrelPos()->getSortenAnfang("startpic");
+//       mSorten.neueSorte(blopart_info, mSorten[gras_nr], false); // kopie vom Gras machen
+//       /* Evtl. verwirrend: Das Gras wurde mit blopart_gras geladen, d.h. die defaults sind auch
+//        * entsprechend... und nicht so, wie man sie für blopart_info erwarten würde */
+//     }
+//        
+      
+
     
     
     /* Ok, alle Sorten geladen. Wenn wir noch mNachbarStapelhoehe zu
@@ -505,27 +551,23 @@ void LevelDaten::ladLevel(int nr) {
     /* Entstehungswahrscheinlichkeiten */
     mKeineGrauenW = daten->getZahlEintragMitDefault("nogreyprob", mVersion, 0);
     if (mKeineGrauenW<0)
-      /* TRANSLATORS: "nogreyprob" is a programming keyword which
-	 should not be translated. */
-      throw Fehler("%s",_("nogreyprob must not be negative"));
+      throw Fehler("%s","nogreyprob must not be negative");
     for (int i=0; i<anzahl_wv; i++) {
       mVerteilungSumme[i]=0;
       for (int j=blopart_min_sorte; j<mAnzFarben; j++)
 	mVerteilungSumme[i] += mSorten[j]->getVerteilung(i);
     }
     if (mVerteilungSumme[wv_farbe]==0)
-      /* TRANSLATORS: Don't worry about the %s, it will be a programming
-	 keyword. */
-      throw Fehler(_("At least one %s must be positive."),
+      throw Fehler("At least one %s must be positive.",
 		   cVerteilungsNamen[wv_farbe]);
     if (mVerteilungSumme[wv_grau]+mKeineGrauenW==0)
-      /* TRANSLATORS: "nogreyprob" is a programming keyword which
-	 should not be translated.
-         Also, don't worry about the %s, it will be a programming keyword. */
-      throw Fehler(_("nogreyprob or at least one %s must be positive."),
+      throw Fehler("nogreyprob or at least one %s must be positive.",
 		   cVerteilungsNamen[wv_grau]);
 
-
+    /* May greys appear? */
+    mGreysAtAll =
+      ((mZufallsGraue>0) || (mSpielerZahl>1)) && (mVerteilungSumme[wv_grau]>0);
+  	  	
    /* Musik (optional) */
    mMusik = daten->getWortEintragMitDefault("music", mVersion, "");
 
@@ -549,8 +591,7 @@ void LevelDaten::ladLevel(int nr) {
     fs += f.getText() + "\n";
     
     if (!mLevelName.isEmpty())
-      // TRANSLATORS: This postpends an error message with some information
-      fs += _sprintf(_(" (Level \"%s\")\n"), mLevelName.data());
+      fs += _sprintf(" (Level \"%s\")\n", mLevelName.data());
     throw Fehler(fs);
   }
   
@@ -598,7 +639,7 @@ int getPunktPos(Str na) {
 /** Liefert zurueck, wie viele Level es gibt. */
 int LevelDaten::getLevelAnz() const {
   if (!mLevelConf[ldteil_summary])
-    throw Fehler("%s",_("Sorry, no working level description file available."));
+    throw Fehler("%s","Sorry, no working level description file available.");
   
   return mIntLevelNamen.size();
 }
@@ -606,7 +647,7 @@ int LevelDaten::getLevelAnz() const {
 /** Lifert zurück, ob die Level der Reihe nach gespielt werden müssen */
 int LevelDaten::getAngeordnet() const {
   if (!mLevelConf[ldteil_summary])
-    throw Fehler("%s",_("Sorry, no working level description file available."));
+    throw Fehler("%s","Sorry, no working level description file available.");
 
   return mAngeordnet;
 }
@@ -643,7 +684,7 @@ Str LevelDaten::getLevelName(int nr) const {
 }
 
 
-/** Liefert die Nummer des Levels mit dem angegebenen Namen zurueck,
+/** Liefert die Nummer des Levels mit dem angegebenen internen Namen zurueck,
     oder 0, wenn der Level nicht existiert. */
 int LevelDaten::getLevelNr(Str na) const {
   /** Alles noch seeehr ineffektiv... */
@@ -655,6 +696,14 @@ int LevelDaten::getLevelNr(Str na) const {
 }
 
 
+bool LevelDaten::levelAccessible(int l) const {
+  if (gDebug || !getAngeordnet()) return true;
+  for (int i=1; i<l; i++)
+    if (!PrefsDaten::getLevelGewonnen(mSpielerZahl == 2, i)) return false;
+  return true;
+}
+
+
 /** Wenn eine Sorte ihre PlatzAnzahl rausgefunden hat, teilt sie uns das mit */
 void LevelDaten::neue_PlatzAnzahl(int PlatzAnzahl) {
   if (mPlatzAnzahlMin==PlatzAnzahl_undefiniert) {
@@ -697,9 +746,7 @@ int LevelDaten::liesDistKey(const Str & key) {
     mDistKeyLen=key.length();
   else
     if (mDistKeyLen!=key.length())
-      /* TRANSLATORS: "distkey" is a programming keyword which should not
-	 be translated */
-      throw Fehler(_("distkey \"%s\" does not have length %d as others do."),
+      throw Fehler("distkey \"%s\" does not have length %d as others do.",
 		   key.data(), mDistKeyLen);
 
   switch (key[0]) {
@@ -707,6 +754,8 @@ int LevelDaten::liesDistKey(const Str & key) {
   case '+': return distkey_farbe;
   case '*': return distkey_gras;
   case '.': return distkey_leer;
+  case '%': return distkey_neighbours;
+  case '&': return distkey_chainreaction;
   default:
     bool anfang = true;
     int n=0;
@@ -726,16 +775,12 @@ int LevelDaten::liesDistKey(const Str & key) {
 	    if (anfang && c==' ')
 	      immernochanfang = true;
 	    else
-	      /* TRANSLATORS: "startdist" and "distkey" are programming keyword
-		 which should not be translated */
-	      throw Fehler(_("Illegal character \"%c\" in startdist or distkey %s"),
+	      throw Fehler("Illegal character \"%c\" in startdist or distkey %s",
 			   c,key.data());
       anfang = immernochanfang;
     }
     if (anfang)
-      /* TRANSLATORS: "startdist" and "distkey" are programming keyword
-	 which should not be translated */
-      throw Fehler("%s",_("All-spaces startdist entry or distkey is not allowed."));
+      throw Fehler("%s","All-spaces startdist entry or distkey is not allowed.");
     return n;
   }
 }
@@ -754,3 +799,14 @@ void LevelDaten::setSchriftFarbe(Color f) {
      and some other pictures */
   AutoColor::gGame = f;
 }
+
+
+/** Liefert true, wenn die angegebene Spalte vom angegebenen Spieler verschoben ist
+    (wegen Hex-Modus) */
+bool LevelDaten::getHexShift(bool rechts, int x) const {
+  if (!ld->mSechseck) return false;
+  else {
+    bool flip = rechts ? (mSechseckFlip & 2) != 0 : (mSechseckFlip & 1) != 0;
+    return (x & 1) != flip;
+  }
+}
