@@ -4,6 +4,11 @@
     begin                : Fri Jul 21 2000
     copyright            : (C) 2000 by Immi
     email                : cuyo@pcpool.mathematik.uni-freiburg.de
+
+Modified 2001-2003,2005,2006,2008,2010,2011,2014 by the cuyo developers
+Modified 2012 by Bernhard R. Link
+Maintenance modifications 2012,2014 by the cuyo developers
+
  ***************************************************************************/
 
 /***************************************************************************
@@ -18,25 +23,29 @@
 #ifndef LEVELDATEN_H
 #define LEVELDATEN_H
 
-#include "inkompatibel.h"
+#include <vector>
+#include <set>
 
-#include <qcolor.h>
-#include <qpixmap.h>
-#include <qstring.h>
-#include <qimage.h>
-#include <qvector.h>
+#include "sdltools.h"
 
 #include "sorte.h"
 #include "bilddatei.h"
+#include "version.h"
+
+class Str;
+class Code;
+class Blop;
+class DefKnoten;
+class ListenKnoten;
+class DatenDatei;
+class Fehler;
 
 
 /***** Konstanten, die irgend was einstellen *****/
 
-/* Gre des Levels */
-#define grx 10
-#define gry 20
+/* Konstanten, die mit dem Layout zusammenhngen: siehe layout.h */
 
-#define platzbild_anzahl 8 // Anzahl der Bildchen der Platz-Animation
+#define max_spielerzahl 2
 
 #define max_farben_zahl 200 // Max. Anzahl der Farben in einem Level
 
@@ -47,45 +56,87 @@
 #define punkte_fuer_kettenreaktion 10
 
 /* Pixel pro Schritt: Bonus-Hetzrand-Geschwindigkeit */
-#define bonus_geschwindigkeit 16
+#define bonus_geschwindigkeit 32
 /* Punkte pro Zeitschritt der Bonus-Animation */
-#define punkte_fuer_zeitbonus 5
+#define punkte_fuer_zeitbonus 10
 
 
 
 /***** Konstanten, die einfach nur was bedeuten *****/
 
-#define zufallsgraue_keine -1
-#define PlatzAnzahl_undefiniert -1
+#define zufallsgraue_keine (-1)
+#define PlatzAnzahl_undefiniert (-1)
 
 
 /* Zur bergabe an laden(), um den Titel-Level zu laden */
-#define level_titel -1
+#define level_titel (-1)
 
 
-/* Fr das Array mSchriftfarbe */
-#define schrift_dunkel 0
-#define schrift_normal 1
-#define schrift_hell 2
+/* Fr die codes in startdist */
+#define distkey_leer (-1)
+#define distkey_gras (-2)
+#define distkey_grau (-3)
+#define distkey_farbe (-4)
+#define distkey_neighbours (-5)
+#define distkey_chainreaction (-6)
+#define distkey_undef (-7)
 
 /* nachbarschaft_... steht in sorte.h */
 
 
-class Code;
-class Blop;
-class DefKnoten;
-class ListenKnoten;
-class DatenDatei;
-class Fehler;
+
+/* Zur Unterscheidung von lazy und eager */
+
+enum LDTeil {
+  ldteil_summary,
+  ldteil_level,
+  ldteile_anzahl
+};
+
+
+
+/** Array, das alle Sorten eines Levels enthlt. Das Hauptfeature dieser Klasse ist, dass
+    Sorten automatisch gelscht werden knnen, und zwar:
+    - nur die, die auch ins Array geschrieben wurden
+    - nur die, die nicht Kopien sind. */
+
+class SortenArray {
+  /** Die eigentliche Liste der Sorten (alles indizes sind um blopart_min_sorte verschoben) */
+  Sorte * mSorten[max_farben_zahl - blopart_min_sorte];
+  /** Welche der mSorten-Pointer sind Kopien von anderen Pointern, und
+   *  welche sind Originale? (Nur originale mssen gelscht werden) */
+  bool mOriginale[max_farben_zahl - blopart_min_sorte];
+
+  /** Laenge des positiven Teils des Arrays; -1 = array leer */
+  int mLaenge;
+  
+ public:
+  SortenArray();
+   
+  /** Sorte auslesen */
+  Sorte * operator[](int nr) const;
+  
+  /** Lnge des Arrays ndern (genauer gesagt: Lnge des positiven Teils);
+   *  sollte nur aufgerufen werden, wenn das Array leer ist. */
+  void setLaenge(int l);
+  
+  /** Sorte in das Array einfgen; normalerweise wird die Sorte dann
+      von loeschen() gelscht; aber nicht, wenn istOriginal = false ist;
+      dann wird davon ausgegangen, dass der Pointer nur eine Kopie ist. */
+  void neueSorte(int nr, Sorte * s, bool istOriginal = true);
+  
+  void loeschen();
+}; // class SortenArray
+
 
 
 /** (Einziges globales) Objekt enthlt alle Informationen ber
     den aktuellen Level.
     
-    Das Parsen der level.descr:
-    - Beim Aufruf des constructors von LevelDaten() wird ladLevelConf()
+    Das Parsen der ld-Dateien:
+    - Beim Aufruf des constructors von LevelDaten() wird ladLevelSummary()
       aufgerufen.
-    - ladLevelConf() sucht die richtige Datei und erzeugt ein
+    - ladLevelSummary() bzw ladLevel  sucht die richtige Datei und erzeugt ein
       DatenDatei-Objekt
     - Der Constructor davon ffnet die Datei und ruft parse() auf.
     - parse() ruft den bison-Parser auf. Der erzeugt folgendes:
@@ -103,30 +154,46 @@ class Fehler;
 
 
 class LevelDaten {
-  friend Sorte::Sorte(__String name, int blopart);
+  friend class Sorte;
+  friend int yyparse();
 	
  public:
-  LevelDaten();
+  LevelDaten(const Version & version);
   ~LevelDaten();
   
-  /** Ld die Levelconf (neu). Wird vom Konstruktor aufgerufen. Und
-      vielleicht im Debug-Modus ab und zu... */
-  void ladLevelConf();
+  /** Ld die Levelconf (neu). Wird vom Konstruktor aufgerufen.
+      Und, wenn sich Einstellungen verndert haben.
+      Eine Vorbedingung ist also, da niemand grad auf irgendwelche Daten
+      dieses Objekts angewiesen ist, insbesondere also, da grad kein
+      Spiel luft.
+      Bei aufJedenFall=false wird nur dann wirklich neu geladen,
+      wenn version!=mVersion.
+      Bei ldteil=ldteil_summary wird am Ende noch der inhalt aller in
+      global= aufgefhrten Dateien nach ldteil_level geladen. */
+  void ladLevelSummary(bool aufJedenFall, const Version & version);
+
+ protected:
+  /** Ld den level-spezifischen Teil */
+  void ladLevelConfig();
+ public:
   
-  /** Wird whrend des Parsens (d. h. innerhalb von ladLevelConf() von
+  /** Wird whrend des Parsens (d. h. innerhalb von ladLevel*()) von
       DefKnoten aufgerufen, wenn ein neuer Level gefunden wurde. Fgt
-      den Level in die Liste der Level ein. ladLevelConf() kann sich
+      den Level in die Liste der Level ein. ladLevelSummary() kann sich
       danach immernoch entscheiden, ob es die Liste wieder lscht und
       durch die "level=..."-Liste ersetzt. */
-  void levelGefunden(__String lna);
+  void levelGefunden(Str lna);
 
-  /** fllt alle Daten in diesem Objekt fr Level nr aus; throwt bei Fehler */
-  void ladLevel(int nr, int spz);
-
-  /** Fllt das von den Daten aus, was auch ohne Level notwendig ist
-      (Punkte-Icons, Schriftfarbe) */
-  void ladKeinenLevel();
+  /* Gibt Speicher frei */
+  void entladLevel();
   
+  /** fllt alle Daten in diesem Objekt fr Level nr aus; throwt bei Fehler */
+  void ladLevel(int nr);
+
+  /** Darf nur aufgerufen werden, wenn nr schon der letzte Level war.
+      Erneuert alles, was seitdem vergessen worden sein knnte. */
+  void erneuerLevel(int nr);
+
   /** Sollte am Anfang des Levels aufgerufen werden; kmmert sich
       um den Global-Blop */
   void startLevel() const;
@@ -139,30 +206,57 @@ class LevelDaten {
   /** Liefert zurck, wie viele Level es gibt. */
   int getLevelAnz() const;
 
+  /** Lifert zurck, ob die Level der Reihe nach gespielt werden mssen */
+  int getAngeordnet() const;
+
   /** Liefert den Namen von Level nr zurck. Liefert "???" bei Fehler. */
-  __String getLevelName(int nr) const;
+  Str getLevelName(int nr) const;
 
   /** Liefert den internen Namen von Level nr zurck. */
-  __String getIntLevelName(int nr) const;
+  Str getIntLevelName(int nr) const;
 
-  /** Liefert die Nummer des Levels mit dem angegebenen Namen zurck,
+  /** Liefert die Nummer des Levels mit dem angegebenen internen Namen zurck,
       oder 0, wenn der Level nicht existiert. */
-  int getLevelNr(__String na) const;
+  int getLevelNr(Str na) const;
+
+  bool levelAccessible(int l) const;
+
+  /** Wenn eine Sorte ihre Platzanzahl rausgefunden hat,
+      teilt sie uns das mit */
+  void neue_PlatzAnzahl(int);
+
+  int zufallsSorte(int wv);
+
+  int liesDistKey(const Str &);
+
+  const Version & getVersion() const;
+
+  /** Setzt AutoColor::gGame */
+  void setSchriftFarbe(Color f);
+  
+  /** Liefert true, wenn die angegebene Spalte vom angegebenen Spieler verschoben ist
+      (wegen Hex-Modus) */
+  bool getHexShift(bool rechts, int x) const;
 
  protected:
- 
-  /** Setzt mSchriftFarbe[...]. Berechnet also insbesondere die dunkle
-      und die helle Farbe. */
-  void setSchriftFarbe(QColor f);
- 
- 
- /** Das Objekt, mit dem man auf die Datei zugreift. Hier steht nur ein
+   
+  /** Ld ein paar Sorten. Wird mehrfach von ladLevel() aufgerufen. */
+  void ladSorten(const Str & ldKeyWort, int blopart);
+   
+ /** Die Objekte, mit denen man auf die Dateien zugreift. Hier stehen nur
      Pointer drauf, damit diese .h-Datei nicht so viel includen muss. */
-  DatenDatei * mLevelConf;
-  /** True, wenn zur Zeit was geladen ist. */
-  bool mGeladen;
+  DatenDatei * mLevelConf[ldteile_anzahl];
+  /** True, wenn zur Zeit der entsprechende Teil der LevelConf geladen ist. */
+  bool mLCGeladen[ldteile_anzahl];
+  /** Alle Level, die schon mal geladen wurden und daher noch vorhanden sind.
+      Hier stehen die Dateinamen. */
+  std::set<Str> mLevelCache;
   /** Liste der internen Levelnamen. */
-  QVector<__String> mIntLevelNamen;
+  std::vector<Str> mIntLevelNamen;
+  /** True falls die Level der Reihe nach gespielt werden mssen. */
+  bool mAngeordnet;
+
+  Version mVersion;
   
  public:
 
@@ -172,31 +266,34 @@ class LevelDaten {
 
 
   /***** Allgemeines *****/
+  bool mLevelGeladen;
+  int mLevelNummer; /* Nummer des geladenen Levels */
   int mSpielerZahl;
   /** Interner Level-Name vom aktuellen Level. (Wird von Aufnahme
       bentigt.) */
-  __String mIntLevelName;
-  __String mLevelName;
-  __String mLevelAutor;
+  Str mIntLevelName;
+  Str mLevelName;
+  Str mLevelAutor;
   /** Der Knoten zum aktuellen Level. */
   DefKnoten * mLevelKnoten;
   /** Beschreibungstext fr den Level */
-  __String mBeschreibung;
-  QColor mHintergrundFarbe;
+  Str mBeschreibung;
+  Color mHintergrundFarbe;
   bool mMitHintergrundbildchen;
   Bilddatei mHintergrundBild;
   /** Farbe der Schrift in dem Level:
       0 = abgedunkelt, 1 = normal, 2 = aufgehellt */
-  QColor mSchriftFarbe[3];
+  Color mSchriftFarbe;
   bool mGrasBeiKettenreaktion;
+  bool mFallPosZufaellig;
   Bilddatei mExplosionBild;
   int mPlatzAnzahlDefault;
   int mPlatzAnzahlMin;
   int mPlatzAnzahlMax;
+  /* Gibt an, ob neben mPlatzAnzahlMin und mPlatzAnzahlMax noch andere
+     PlatzAnzahlen vorkommen. */
   bool mPlatzAnzahlAndere;
-  Bilddatei mPktZiffernBild[2];
-  Bilddatei mDbgZiffernBild;
-  __String mMusik;
+  Str mMusik;
   
   /** Max. Anzahl der Bilder, die ein Blop gleichzeitig malt. Wird
       (ggf.) von den Sorten erhht, wenn man sie ldt. */
@@ -208,27 +305,24 @@ class LevelDaten {
   int mNachbarStapelHoehe;
   
   /***** Die Sorten *****/
- protected:
-  /** Das Array, in dem alle Sorten stehen. Allerdings soll die Index-
-      Menge nicht bei 0 losgehen, sondern bei blop_min_sorte. Deshalb
-      gibt es in public die Variable mSorten... */
-  Sorte * mSortenIntern[max_farben_zahl - blopart_min_sorte];
  public:
-  /** Wird ganz am Anfang auf mSortenIntern - blopart_min_sorte
-      initialisiert. */
-  Sorte * * const mSorten;
+  /** Die ganzen Sorte-Objekte */
+  SortenArray mSorten;
   int mAnzFarben;
-  int mAnzFarbnamen;
- 
+  int mVerteilungSumme[anzahl_wv];
+  int mKeineGrauenW;    /** Zhlt nicht zu mGrauSumme */
+
   /***** Hetzrand *****/
-  QColor hetzrandFarbe;
+  Color hetzrandFarbe;
   int hetzrandZeit;
   bool mMitHetzbildchen;
   Bilddatei mHetzBild;
   int mHetzrandUeberlapp;
+  int mHetzrandStop;
 
   /***** Gras *****/
   ListenKnoten * mAnfangsZeilen;
+  int mDistKeyLen; /** Lnge der distKeys. 0, wenn es noch keinen gab. */
 
 
   /***** KI-Player-Nutzen-Funktion *****/
@@ -250,12 +344,15 @@ class LevelDaten {
   int mNachbarschaft; // Default-Wert fr die Sorten
   /** true bei Sechseckraster. Wird direkt aus mNachbarschaft bestimmt. */
   bool mSechseck;
+  /* Nur fuer Sechseckraster: Welche(r) Spieler ist links-rechts gespiegelt? */
+  int mSechseckFlip;
   bool mMitLeerBildchen;
   int mZufallsGraue;
-  	
- protected: // Protected methods
-  /** Throwt, wenn der angegebene Schlssel in level.descr nicht existiert. */
-  void verlangeSchluessel(const char * s) const;
+  bool mGreysAtAll;
+
+ protected:
+
+  bool mSammleLevel; /* Bestimmt, ob levelGefunden berhaupt was tun soll. */
 
 };
 
