@@ -4,6 +4,9 @@
     begin                : Thu Jul 26 2001
     copyright            : (C) 2001 by Immi
     email                : cuyo@karimmi.de
+
+Modified 2001,2002,2005,2006,2011,2014 by the cuyo developers
+
  ***************************************************************************/
 
 /***************************************************************************
@@ -18,8 +21,6 @@
 #ifndef NACHBARITERATOR_H
 #define NACHBARITERATOR_H
 
-#include "inkompatibel.h"
-
 /**Iteriert durch alle Nachbarn eines Punktes.
 (Nachbarn im Sinne von Kette zhlt als verbunden.)
 Verwendung:
@@ -33,7 +34,7 @@ class Sorte;
 
 class NachbarIterator {
 public: 
-  NachbarIterator(const Sorte * s, int x, int y);
+  NachbarIterator(const Sorte * s, bool rechts, int x, int y);
   ~NachbarIterator();
 
   /** Nchster Nachbar */
@@ -46,6 +47,7 @@ public:
 protected:
 
   int mNachbarschaft;
+  bool mRechts;
   int mX0, mY0;
   int mI;
   bool mEnde;
