@@ -4,6 +4,9 @@
     begin                : Wed Jul 12 2000
     copyright            : (C) 2000 by Immi
     email                : cuyo@pcpool.mathematik.uni-freiburg.de
+
+Modified 2001-2007,2010,2011,2014 by the cuyo developers
+
  ***************************************************************************/
 
 /***************************************************************************
@@ -18,20 +21,17 @@
 #ifndef SPIELFELD_H
 #define SPIELFELD_H
 
-#include "inkompatibel.h"
-
 #include <cstdlib>
 
-#include <qwidget.h>
-#include <qtimer.h>
-
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
 
-#include "blop.h"
+#include <SDL.h>
+
 #include "blopgitter.h"
 #include "fall.h"
+#include "font.h"
 
 
 #define graue_bei_kettenreaktion 5 // (grx - 1) // (grz - minflopp)
@@ -40,8 +40,6 @@
    verschiebt; sollte Teiler von gric sein */
 #define reihe_rueber_senkrecht_pixel 8
 
-#define fall_langsampix 6
-
 
 
 /* Fr das Grafik-Update-System... */
@@ -59,22 +57,25 @@
 
 
 
+enum {
+  infoblop_grey,
+  infoblop_grass,
+  infoblop_neighbours,
+  infoblop_chainreaction,
+  infoblop_anz
+};
+
 
 class KIPlayer;
 
 /**
-  *@author immi
+  *
   */
 
-class Spielfeld : public QWidget  {
-  Q_OBJECT
+class Spielfeld {
 public:
-  Spielfeld(bool re, QWidget *parent=0, const char *name=0);
+  Spielfeld(bool re);
   ~Spielfeld();
-  /** Sollte einmal direkt nach dem Start von cuyo aufgerufen werden;
-      initialisiert den aktuellen Level und benutzt das Startbild als
-      Titelbild. */
-  void erzeugTitelbild();
   /** sollte nur aufgerufen werden, wenn papa->spielLaeuft() false
       liefert; liefert
       true, wenn alle Spiel-Stop-Animationen fertig sind; liefert
@@ -87,15 +88,18 @@ public:
   /** Liefert true, wenn grade ein Fallendes unterwegs ist.
       Wird vom KIPlayer bentigt */
   int getFallModus() const;
-  /** Liefert true, wenn noch Gras da ist (d. h. wenn nicht gewonnen.) */
-  bool istGrasDa() const;
   /** Liefert einen Pointer auf das Blopgitter zurck. */
   BlopGitter * getDatenPtr();
   /** Liefert einen Pointer auf die fallendne Blops zurck.
       Wird von KIPlayer einmal am Anfang aufgerufen
-      und gelegentlich bei @0 und  @1. */
-  const Blop * getFall() const;
-  Blop * getFall();
+      und von ort_absolut::finde(). */
+  const Blop * getFall(int which = 0) const;
+  Blop * getFall(int which = 0);
+  int getFallAnz() const;
+  /** Ersetzt das Fall durch ein neues, das wieder von ganz oben startet.
+      Aber nur, wenn noch beide Hlften dran sind. */
+  void resetFall();
+  Blop & getSemiglobal();
   /** Liefert die Pos. zurck, an der neue Dinge oben
       auftauchen. */
   int getHetzrandYAuftauch() const;
@@ -105,85 +109,74 @@ public:
   /** Liefert die Pos. vom Hetzrand in Pixeln zurck. Wird
       vom Fall gebraucht. */
   int getHetzrandYPix() const ;
+  /** ndert die Hhe vom Hetzrand auf y (in Pixeln). */
+  void setHetzrandYPix(int y);
   /** Liefert die (rberreihenbedingte) Hochverschiebung
       des gesamten Spielfelds. Wird vom Fall bentigt (um
       seine Koordinaten in Feldern zu berechnen). */
-  int getHochVerschiebung();
-  /** setzt den Bereich vom Fallenden auf upzudaten */
-  void setUpdateFall(int ebene = ebene_hintergrund);
+  int getHochVerschiebung() const;
+  /** Liefert die Koordinaten eines Felds in Pixeln zurck (ungespiegelt) */
+  void getFeldKoord(int x, int y, int & xx, int & yy) const;
 
-  /** Zeigt t gro an. (Oder weniger gro.) Auf Wunsch wird
-      ein Update-Event gesendet. (Sonst ist der Aufrufer selbst
-      dafr verantwortlich, dass der Text auch wirklich mal
-      angezeigt wird.) */
-  void setText(__String t, bool update_event = true,
-               bool kleine_schrift = false);
+  /** Zeigt t gro an. (Oder weniger gro.) */
+  void setText(const Str & t, bool kleine_schrift = false,
+	       int x0 = 0, int x1 = gric*grx);
   
   /** Fr whrend des Spiels: Setzt einen Text, der ein paar mal
       aufblinkt. */
-  void setMessage(__String mess);
+  void setMessage(Str mess);
+
+  /** Setzt das Rechteck x, y, w, h auf upzudaten. */
+  void setUpdateRect(int x, int y, int w, int h);
+  /** Setzt alles auf upzudaten. */
+  void setUpdateAlles();
+  /** Setzt das nchste Fall auf upzudaten... und die Infblops */
+  void setUpdateNaechstesFall();
+
+  /** Hauptmalroutine: Malt alles, was sich geaendert hat, neu */
+  void malUpdateAlles();
+  /** Malt (falls noetig) das nchste Fall neu... und die Infoblops */
+  void malUpdateNaechstesFall();
+
+  int getZeit() const;
+  
+  /** Liefert true, wenn diese Spalte um gric/2 verschoben ist (wegen Hex-Modus) */
+  bool getHexShift(int x) const;
+ protected:
+  /** Malt die ganzen Blops, usw. */
+  void malUpdateSpielsituation();
+  /** Malt die Schrift auf den Bildschirm */
+  void malSchrift(bool mitte /*= true*/,
+                  const FontStr & text,
+		  int x0, int x1);
+  /** Wenn der Level gespiegelt ist, wird auch das Rechteck gespiegelt. */
+  void spiegelRect(SDL_Rect & r);
 
- protected: // Protected methods
   /** liefert die Hhe vom hchsten Trmchen */
   int getHoehe();
-  /**  */
-  void paintEvent(QPaintEvent *);
-  /** Liefert die Koordinaten eines Felds in Pixeln zurck (ungespiegelt) */
-  QPoint getFeldKoord(int x, int y) const;
-  /** liefert eine VerbindungsBitliste fr den Blop bei x, y. Bezieht sich
-      ggf. auf's Feuer. */
-  int getVerbindungen(int x, int y, bool feuer = false) const;
   /** berechnet, ob und welche Blops platzen mssen. Auerdem
       werden den Blops die Kettengren mitgeteilt, und Graue
       und Punkte verteilt */
   bool calcFlopp();
   /** sucht die Zusammenhangskomponente von x, y mit Farbe n.
-      Setzt flopp[][] an den entspr. Stellen auf w.
-      Aufruf mit w = -1 um rauszufinden, ob genug gleichfarbige beisammen sind;
+      Setzt in flopp[][] an den entspr. Stellen das Bit w.
+      Aufruf mit w = 1 um rauszufinden, ob genug gleichfarbige beisammen sind;
       liefert Anzahl der gleichfarbigen zurck;
-      Danach Aufruf mit w = 1 und richtigem anz-Wert, um den Blops ihre neue
+      Danach Aufruf mit w = 2 und richtigem anz-Wert, um den Blops ihre neue
       Kettengre mitzuteilen.
       Liefert Anzahl der Punkte zurck, die's dafr gibt (wenn Blops platzen).
-      Wenn auch_gras true ist, wird auch benachbartes Gras gesprengt.
       Nur wenn platzen = true ist, wird platzen durchgefhrt.
       Sonst bekommen die
       Blops nur ihre Kettengre mitgeteilt.
       @return Je nach w Gre der Zshgskomp. oder Anz. d. Punkte */
   int calcFloppRec(int flopp[grx][gry], int x, int y,
 		   int n, int w, bool platzen = false,
-		   bool auch_gras = false, int anz = 0);
-  /** Ruft malUpdateIntern() auf und malt das interne Bild dann
-      auf p.
-      Bei force = true wird nicht darauf geschaut, ob die Zeit gerade
-      knapp ist. */
-  void malUpdate(QPainter & p, bool force = false);
-  /** Hier findet die Grafik statt: Alles was in den mUpdateXXX-
-      Variablen angegeben ist (auch in denen von den Blops selbst),
-      wird auf den internen Puffer neu gezeichnet. */
-  void malUpdateIntern();
+		   bool ist_kettenreaktion = false, int anz = 0);
   /** verschiebt einen Blop (auch wenn er explodiert oder sonstwie
       grad animiert ist). Was dort hin soll, von wo der Blop weg ist
       (z. B. blopart_wirklich_keins) kann bergeben werden. */
   void verschiebBlop(int x1, int y1, int x2, int y2,
 		     int bg_sorte = blopart_keins);
-  /** ndert die Hhe vom Hetzrand auf y (in Pixeln).
-      Insbesondere wird auch hetzrandY gendert. */
-  void setHetzrandYPix(int y);
-  /** Wenn der Level gespiegelt ist, wird auch r gespiegelt */
-  QRect spiegelRect(const QRect & r);
-  /** Setzt das Rechteck r auf upzudaten. Der ebene-Parameter sollte
-      (im Moment) nur von malUpdate() benutzt werden. */
-  void setUpdateRect(QRect r, int ebene = ebene_hintergrund);
-  /** Setzt das Rechteck x, y, w, h auf upzudaten. */
-  void setUpdateRect(int x, int y, int w = gric, int h = gric,
-		     int ebene = ebene_hintergrund);
-  /** Setzt alles auer den Hetzrand auf upzudaten. */
-  void setUpdateFastAlles();
-  /** Malt die Schrift auf den Bildschirm */
-  void malSchrift(QPainter & p, bool mitte = true);
-  /** Lsst ein neues Fall oben auftauchen. Liefert false, wenn
-      dafr kein Platz ist. */
-  bool neuesFall();
   /** Lsst in der Luft hngende Blops ein Stck runterfallen.
       Liefert zurck, ob sich nichts bewegt hat, nur unten oder auch
       oben. Bei auchGraue = true, kommen auch Graue, die ganz ber
@@ -209,27 +202,29 @@ public:
   /** true, wenn rechter von den beiden Spielern */
   bool mRechterSpieler;
 
-  /** Das Titelbild (fr whrend kein Spiel luft) */
-  QPixmap mTitelbild;
-	
   /***** Variablen, die sich hufig ndern *****/
 
-  /** Kopie vom Bildschirm fr flackerfreies Malen
-      (nur vom Spielgeschehen, nicht von Schrift, die drber steht;
-      also von dem, was malUpdate() / malUpdateIntern() malt) */
-  QPixmap mBild;
-	
   /** Text, der grade angezeigt wird (falls das Spiel nicht luft) */
-  __String mText;
+  FontStr mText;
   /** True, wenn die Schrift von mText klein sein soll. */
   bool mTextKlein;
+  /** Horizontal clipping coordinates for mText. */
+  int mTextX0,mTextX1;
   /** Text, der grade blinkt */
-  __String mMessageText;
+  Str mMessageText;
   /** Blink-Zeitzhler. */
   int mMessageZeit;
 		
   /** _die_ Spielfelddaten */
   BlopGitter mDaten;
+  /** Alle Fall-Daten (Pointer, damit "mFall = mNaechsterFall" keine Daten rumschieben muss) */
+  Fall *mFall;
+  Fall *mNaechsterFall;
+  /** Der semi-globale Blop */
+  Blop mSemiglobal;
+  /** Die Info-Blops */
+  Blop mInfoBlops[infoblop_anz];
+  bool mInfoBlopActive[infoblop_anz];
   /** Zeit innerhalb des Levels */
   int mZeit;
   /** Hhe in Pixeln vom Rand, der mit der Zeit runterkommt,
@@ -237,8 +232,6 @@ public:
   int mHetzrandYPix;
   /** Hhe in Feldern vom Rand */
   int mHetzrandY;
-  /** Alle Fall-Daten */
-  Fall mFall;
   /** True, wenn das Fall grade frisch aufgetaucht ist. Der KI-Player will
       das wissen. */
   bool mFallIstNeu;
@@ -257,46 +250,39 @@ public:
   bool mKettenreaktion;
   /** Anzahl der Grauen, die darauf warten, runterzukommen */
   int mGrauAnz;
+  int mGrasAnz;
   /** != 0 whrend des Reihen hin und hergebens. Um so viele Pixel ist das
       gesamte Spielfeld nach oben verschoben. */
   int mHochVerschiebung;
 
-  /** Die nachfolgenden Variablen werden nur whrend einer
-      Rochade-Animation gebraucht */
-  int mRoY;
-  bool mRoGross;
-  int mRoZeit;
-  Blop mRoTurm;
-  Blop mRoKoenig;
-	
-  /***** Variablen, die sich so hufig ndern, dass sie nach einem
-         Funktionsaufruf von auen nicht mehr bentigt werden *****/
-	
-	
-  /* Die folgenden Variablen geben an, was von malUpdate() neu gemalt werden soll.
-     Sie sollten nur von setUpdateRect() gendert werden. */
-  /** Welche Datenfelder sollen geupdatet werden? In die untere Bonuszeile wird im
-      Sechseck-Modus ab und zu evtl. Mll geschrieben. */
-  bool mUpdateDaten[grx][gry + 2];
-  bool mUpdateFall;
-  bool mUpdateHetzrand;	// der Hetzrand (samt Bild)
-  /** Bereiche in den verschiedenen Ebenen, die upgedatet werden mssen.
-      Kann null (statt empty) sein.
-      Achtung: Schon gespiegelt */
-  QRegion mUpdateRegion;
+
+  /* Die folgenden Variablen geben an, was von malUpdate() neu gemalt werden
+     soll. Sie sollten nur von setUpdateRect() u.ae. gendert werden. */
+  bool mUpdateBlop[grx][gry + 2];  // welche einzelnen Blops neu malen (inkl. Hexmodus-Unterrand)
+  bool mUpdateAlles;   // true heisst: alles neu malen
+  bool mUpdateNaechstesFall;
+
+ private:
+  /* Liefert zurueck, zu wie vielen Nachbarblops sich Blop b
+     an Position x,y verbinden wuerde.
+     Wird von createStartDist() verwendet. */
+  int blopVerbindungen(const Blop & b, int x, int y);
+ 
+  /* Startdist aus Leveldaten auslesen und ins Spielfeld schreiben */
+  void createStartDist();
 	
  public:
+  /** Schaltet die Spielfeld-Anzeige aus, damit "Loading level"
+      angezeigt werden kann... */
+  void ladeLevelModus();
   /** Initialisiert alles fr's Spiel. Schaltet die Spielfeldanzeige an.
       Muss in einer Gleichzeit aufgerufen werden, fr die init-Events. 
       Danach muss noch einmal animiere() aufgerufen werden, damit alle
       Blops wissen, wie sie aussehen, und damit die Grafik gemalt wird. */
   void startLevel();
   /** Spiel abbrechen (sofort, ohne Animation; oder die
-      Animation ist schon vorbei).
-      Wenn malen = true ist, Bildschirm sofort updaten.
-      Wenn wegen_fehler = true ist, wird das Spiel trotzdem weiter
-      angezeigt. */
-  void stopLevel(bool malen, bool wegen_fehler = false);
+      Animation ist schon vorbei). */
+  void stopLevel();
   
   
   /***** Funktionen, die einmal pro Spielschritt aufgerufen werden *****/
@@ -322,14 +308,15 @@ public:
   void testeFlopp();
  
   /** Ein Schritt vom Spiel.
-      Animationen und das Grafik-Update werden *nicht* gemacht.
-      Dazu muss animiere() aufgerufen werden. (Weil alle Animationen
-      innerhalb einer Gleichzeit stattfinden sollen.) */
+      Animationen wird *nicht* gemacht. Dazu muss animiere() aufgerufen werden.
+      (Weil alle Animationen innerhalb einer eigenen Gleichzeit
+      stattfinden sollen.)
+      spielSchritt() sollte innerhalb einer Gleichzeit aufgerufen
+      werden fr evtl. auftretende Events. */
   void spielSchritt();
-  /** Fhrt alle Animationen durch, und bei der Gelegenheit auch
-      die Grafikupdates (auch die, die evtl. frher beim
-      spielSchritt()-Aufruf angefallen sind). Sollte innerhalb einer
-      Gleichzeit aufgerufen werden. */
+  /** Fhrt alle Animationen durch.
+      Sollte innerhalb einer Gleichzeit aufgerufen werden.
+      Zählt auch die Gras-Blobs neu durch. */
   void animiere();
   
   
@@ -348,33 +335,19 @@ public:
   /** ndert die Fallgeschwindigkeit vom Fall */
   void tasteFall();
 
- public slots:
+ public:
   /** Graue von anderem Spieler bekommen; wird ignoriert, falls dieser
       Spieler grad nicht spielt */
   void empfangeGraue(int g);
-  /** liefert (in ret) zurck, ob wir dem anderen
-      Spieler eine Reihe geben (er hat Hhe h); eine der Konstanten
-      bewege_reihe_xxx */
-  void gebReihe(int h, int & ret);
+  int getGrauAnz() const;
+  int getGrasAnz() const;
+  /** liefert zurck, ob wir dem anderen
+      Spieler eine Reihe geben (er hat Hhe h); Antwort ist
+      eine der Konstanten bewege_reihe_xxx */
+  int bitteUmReihe(int h);
   /** gibt einen Stein an den anderen Spieler
-      rber; in s wird die Steinfarbe zurckgeliefert */
+      rber; Blop wird in s zurckgeliefert */
   void gebStein(Blop & s);
-
- signals: // Signals
-  /** emit, wenn das Spiel (fr diesen Spieler) zu Ende ist */
-  void sigTot();
-  /** emit, wenn Graue an andere Spieler zu liefern sind (Param = Anz. d. Grauen) */
-  void sigSendeGraue(int);
-  /** emit, wenn evtl. eine Reihe vom anderen Spieler her sollte;
-      der 1. Parameter ist dann die eigene Hhe; im zweiten speichert
-      der andere Spieler die Antwort: eine der Konstanten
-      bewege_reihe_xxx */
-  void sigWillReihe(int, int &);
-  /** emit, wenn ein Stein von der Reihe, die grad rber geht, rbergegeben werden
-      soll; Steinart wird im Parameter gespeichert */
-  void sigWillStein(Blop &);
-  /** emit, wenn dieser Spieler Punkte bekommt. Parameter: rechterSpieler, #Punkte */
-  void sigBekommPunkte(bool, int);
 };
 
 #endif
