@@ -4,6 +4,9 @@
     begin                : Thu Jul 20 2000
     copyright            : (C) 2000 by Immi
     email                : cuyo@pcpool.mathematik.uni-freiburg.de
+
+Modified 2002,2005,2006,2008,2010,2011,2014 by the cuyo developers
+
  ***************************************************************************/
 
 /***************************************************************************
@@ -17,10 +20,9 @@
 
 
 #include "variable.h"
-#include "definition.h"
 #include "blop.h"
 #include "cuyointl.h"
-#include "fehler.h"
+#include "ort.h"
 
 
 /** Erzeugt eine Mll-Variable. Wird verwendet, wenn es einen Fehler
@@ -33,72 +35,42 @@ Variable::Variable():
 }
 
 
-Variable::Variable(//__String datna, int znr, 
+Variable::Variable(//Str datna, int znr, 
                    VarDefinition * d,
-                   int dx, int dy /*= 0*/):
+                   Ort * ort):
   //mDateiName(datna), mZeilenNr(znr),
-  mDeklaration(d), mDX(dx), mDY(dy) {
+  mDeklaration(d), mOrt(ort) {
   
 }
 
 
 
-__String Variable::toString() const {
-  __String ret;
-  if (mDX == variable_global) {
-    ret.sprintf("%s@", getName().data());
-  } else if (mDX == variable_lokal) {
-    ret.sprintf("%s", getName().data());
-  } else {
-    ret.sprintf("%s@(%d,%d)", getName().data(), mDX, mDY);
-  }
-  return ret;
-/*  __String ret2;
-  ret2.sprintf("%s (%s:%d)", ret.data(), mDateiName.data(), mZeilenNr);
-  return ret2;*/
+Variable::~Variable() {
+  if (mOrt) delete mOrt;
+}
+
+
+
+Str Variable::toString() const {
+  Str ret1 = getName();
+  if (mOrt) ret1 = ret1 + mOrt->toString();
+  return ret1;
+
+/*  Str ret2;
+  return _sprintf("%s (%s:%d)", ret.data(), mDateiName.data(), mZeilenNr);
+*/
 }
 
 
 
 
-
-/** Liefert die absoluten Koordinaten, so wie sie intern gebraucht
-    werden (also ggf. gespiegelt und Hex-bereinigt). (y wird von
-    getDX() zwar nicht gebraucht, aber so sieht's schner aus.) */
-int Variable::getX(int x, int /*y*/) const {
-  return x + mDX;
+bool Variable::Ort_hier() const {
+  if (mOrt)
+    return mOrt->hier();
+  else
+    return true;
 }
 
-int Variable::getY(int x, int y) const {
-  /* Hier findet das Spiegeln der relativen Variablenkoordinaten
-     fr den Himmel-Level statt. (User will y nach unten eingeben;
-     intern ist im Himmel-Level aber y nach oben.) */
-  y += (ld->mSpiegeln ? -mDY : mDY);
-
-  /* Und jetzt ggf. die Anpassung an die Hex-Koordinaten.
-     Intern sind die relativen Koordinaten bei ungeradem dx so gespeichert,
-     dass dy = 0 leicht schrg nach oben bedeutet. D. h. Fr die graden
-     Spalten stimmts, fr die ungeraden muss noch verschoben werden.
-  */
-  if (ld->mSechseck) {
-    if ((x & 1) && (mDX & 1))  // Oder: x & mDX & 1...
-      y--;
-  }
-
-  return y;
-}
-
-
-
-/** Liefert bei variable_fall 0 oder 1, je nachdem welcher Fall-Blop
-      gemeint ist. Das wurde dem Konstruktor ber dy mitgeteilt. */
-int Variable::getWelcherFall() const {
-  CASSERT(mDX==variable_fall);
-  CASSERT((mDY==0) || (mDY==1));
-  return mDY;
-}
-
-
 
 int Variable::getNummer() const {
   if (!mDeklaration)
@@ -107,11 +79,11 @@ int Variable::getNummer() const {
 }
 
 
-int Variable::getOrt() const {
-  if (mDX == variable_global || mDX == variable_lokal || mDX == variable_fall)
-    return mDX;
+ort_absolut Variable::getOrt(ort_absolut vonhieraus, Blop & fuer_code) const {
+  if (mOrt)
+    return mOrt->berechne(vonhieraus, fuer_code);
   else
-    return variable_relativ;
+    return vonhieraus;
 }
 
 
@@ -131,9 +103,9 @@ int Variable::getDefaultWert() const {
   return mDeklaration->mDefault;
 }
 
-__String Variable::getName() const {
+Str Variable::getName() const {
   if (!mDeklaration)
-    return _("<unknown>");
+    return "<unknown>";
   return mDeklaration->mName;
 }
 
