@@ -4,6 +4,9 @@
     begin                : Fri Apr 20 2001
     copyright            : (C) 2001 by Immi
     email                : cuyo@karimmi.de
+
+Modified 2002,2003,2006,2008-2011,2014 by the cuyo developers
+
  ***************************************************************************/
 
 /***************************************************************************
@@ -18,6 +21,8 @@
 #include <cstdlib>
 #include <cstdio>
 
+#include "sdltools.h"
+
 #include "cuyointl.h"
 #include "fehler.h"
 #include "xpmladen.h"
@@ -58,7 +63,7 @@ void gzclose(int) {}
 
 /** Wird von ladeDatei benutzt, um entweder eine normale oder eine
     gz-Datei zu laden... */
-bool ladeDateiLow(__String na, bool gz) {
+bool ladeDateiLow(Str na, bool gz) {
   FILE * f = 0; // Das "= 0" ist nur um keine Warnungen zu bekommen
   gzFile gzf = 0;
 
@@ -88,15 +93,14 @@ bool ladeDateiLow(__String na, bool gz) {
 
       if (neu < 0) {
         gzclose(gzf);
-	/* Hier knnte man schnere Fehlermeldungen mit gzerror()
-	   bekommen... */
-        throw Fehler(_("Read error"));
+        /* gzerror() could improve the error message... */
+        throw Fehler("%s","Read error");
       }
     } else {
       bei += fread(gDatAnfang + bei, 1, neu, f);
       if (ferror(f)) {
         fclose(f);
-        throw Fehler(_("Read error"));
+        throw Fehler("%s","Read error");
       }
     }
   }
@@ -124,7 +128,7 @@ bool ladeDateiLow(__String na, bool gz) {
     an.
     Liefert false, wenn die Datei nicht gefunden wird.
     Sucht auch nach der Datei na.gz. */
-bool ladeDatei(__String na) {
+bool ladeDatei(Str na) {
 
   /* Erst mal versuchen, eine ungezippte Datei zu laden. */
   if (ladeDateiLow(na, false)) return true;
@@ -157,7 +161,7 @@ void leerUndKommentarWeg() {
     gDatBei += 2;
     while (gDatBei[0] != '*' || gDatBei[1] != '/') {
       if (!gDatBei[0])
-        throw Fehler(_("Endless comment."));
+        throw Fehler("%s","Endless comment.");
       gDatBei++;
     }
     gDatBei += 2;
@@ -165,11 +169,11 @@ void leerUndKommentarWeg() {
 }
 
 /** Erwartet, dass ein s kommt (davor ist whitespace erlaubt) */
-void erwarte(char * s) {
-  char *t = s;
+void erwarte(const char * s) {
+  const char *t = s;
   while (*s) {
     if (*gDatBei != *s)
-      throw Fehler(_("\"%s\" expected"), t);
+      throw Fehler("\"%s\" expected", t);
     gDatBei++;
     s++;
   }
@@ -179,7 +183,7 @@ void erwarte(char * s) {
 /** Das gleiche mit char */
 void erwarte(char c) {
   if (*gDatBei != c)
-    throw Fehler(_("'%c' expected; found: '%c' (filepos: %d)"), c, *gDatBei,
+    throw Fehler("'%c' expected; found: '%c' (filepos: %d)", c, *gDatBei,
               (int) (gDatBei - gDatAnfang));
   gDatBei++;
 }
@@ -190,7 +194,7 @@ void erwarte(char c) {
 void liesBis(char c) {
   while (*gDatBei != c) {
     if (*gDatBei == 0)
-      throw Fehler(_("'%c' expected"), c);
+      throw Fehler("'%c' expected", c);
     gDatBei++;
   }
   gDatBei++;
@@ -199,7 +203,7 @@ void liesBis(char c) {
 
 /** Prft, ob der String s jetzt kommt. Wenn ja, wird
     er weggelesen */
-bool kommtString(char * s) {
+bool kommtString(const char * s) {
   char * merk = gDatBei;
   while (*s) {
     if (*gDatBei != *s) {
@@ -224,7 +228,7 @@ int getInt() {
     geht = true;
   }
   if (!geht)
-    throw Fehler(_("Number expected"));
+    throw Fehler("%s","Number expected");
   return ret;
 }
 
@@ -233,9 +237,11 @@ int getInt() {
 int decodeHex1(char a) {
   if (a >= '0' && a <= '9')
     return a - '0';
-  if (a >= 'A' && a <= 'F' || a >= 'a' && a <= 'f')
-    return ((a - 'A') & 7) + 10;
-  throw Fehler(_("Hex number expected"));
+  if (a >= 'A' && a <= 'F')
+    return a - 'A' + 10;
+  if (a >= 'a' && a <= 'f')
+    return a - 'a' + 10;
+  throw Fehler("%s","Hex number expected");
 }
 
 int getHex() {
@@ -245,34 +251,81 @@ int getHex() {
 
 
 
-/** Falls die eigene Lad-Routine nicht funktioniert, die
-    von QT probieren. */
-bool fallBackLaden(QImage & im, __String na) {
-  if (gDebug)
-    fprintf(stderr, _("Using (slow) QT-Routine.\n"));
-  return im.load(na);
-}
+
+struct info {
+  Uint32 farbcode;
+  bool hintergrund;
+};
+
+union suchbaum {
+  info einziges;
+  info blatt[256];
+  suchbaum * weiter[256];
+
+  suchbaum(int tiefe) {
+    if (tiefe>1)
+      for (int i=0; i<256; i++) weiter[i]=NULL;
+  }
+
+  inline info* rein(int tiefe) {
+    if (tiefe==0)
+      return &einziges;
+    suchbaum* such = this;
+    for (;tiefe-->1;) {
+      if (!such->weiter[(unsigned int)*gDatBei])
+        such->weiter[(unsigned int)*gDatBei] = new suchbaum(tiefe);
+      such = such->weiter[(unsigned int)*gDatBei++];
+    }
+    return (such->blatt) + (*gDatBei++);
+  }
+
+  inline info raus(int tiefe) {
+    if (tiefe==0)
+      return einziges;
+    suchbaum* such = this;
+    for (;tiefe-->1;) {
+      if (!such->weiter[(unsigned int)*gDatBei])
+        throw Fehler("%s","Undefined pixel name");
+      such = such->weiter[(unsigned int)*gDatBei++];
+    }
+    return such->blatt[(unsigned int)*gDatBei++];
+  }
+
+  void loesch(int tiefe) {
+    if (tiefe-->1)
+      for (int i=0; i<256; i++)
+        if (weiter[i]) {
+	  weiter[i]->loesch(tiefe);
+          delete weiter[i];
+        }
+  }
+};
 
 
-/* Versucht die Datei na zu laden und in QImage zu speichern.
-   Versucht auerdem, die Datei na.gz zu lagen.
-   Liefert false, wenn keine der Dateien existiert.
+
+
+
+
+/* Versucht die Datei na zu laden.
+   Versucht auerdem, die Datei na.gz zu laden.
+   Liefert 0, wenn keine der Dateien existiert.
    Throwt, wenn's beim Laden einen Fehler gibt.
-   (Falls die QT-Lad-Routine verwendet wird, kann nicht versucht werden,
+   (Falls die SDL-Lad-Routine verwendet wird, kann nicht versucht werden,
    die .gz-Datei zu laden.) */
-bool ladXPM(QImage & im, __String na) {
+SDL_Surface * ladXPM(Str na, RohMaske & maske) {
+  SDL_Surface * s;
 
   gDatAnfang = 0;
   
   gDateiGezippt = false;
   
   /* Das nachfolgende try-catch ist 1. um evtl Speicher freizugeben
-     und zweitens, weil wir es dann nochmal mit der QT-Laderoutine
+     und zweitens, weil wir es dann nochmal mit der SDL-Laderoutine
      versuchen wollen. */
   try {
 
     /* Datei laden. Dabei werden gDatAnfang und gDatEnde gesetzt. */
-    if (!ladeDatei(na)) return false;
+    if (!ladeDatei(na)) return 0;
 
 
     gDatBei = gDatAnfang;
@@ -292,75 +345,94 @@ bool ladXPM(QImage & im, __String na) {
     leerUndKommentarWeg();
     erwarte('"');
 
-    int groesse_x, groesse_y, farb, charpp;
+    int groesse_x, groesse_y, farb, tiefe;
     groesse_x = getInt();
     groesse_y = getInt();
     farb = getInt();
-    charpp = getInt();
-    if (charpp != 1)
-      throw Fehler(_("The pixels consist of several chars. That's not yet supported in .xpm.gz files."));
+    tiefe = getInt();
+    bool monochrom = tiefe==0;
     leerWeg();
     erwarte('"');
 
-    im.create(groesse_x, groesse_y, 32);
-    im.setAlphaBuffer(true);
+    s = SDLTools::createSurface32(groesse_x, groesse_y);
+    SDL_PixelFormat * pf = s->format;
 
+    maske.init(groesse_x,groesse_y);
 
     /* Farben parsen */
 
-    QRgb farben[256];
+    suchbaum farben(tiefe);
+
     for (int i = 0; i < farb; i++) {
       leerUndKommentarWeg();
       erwarte(',');
       leerUndKommentarWeg();
       erwarte('"');
-      
-      int index = *gDatBei++;
+
+      info* index = farben.rein(tiefe);
       leerWeg();
       char typ = *gDatBei++;
       CASSERT(typ == 'c');
       leerWeg();
-            
+
+      index->hintergrund=false;
       if (kommtString("None")) {
         // durchsichtig
-	farben[index] = 0;
+	index->farbcode = SDL_MapRGBA(pf, 0, 0, 0, 0);
       } else if (kommtString("black")) {
-        farben[index] = qRgb(0, 0, 0) | 0xff000000;
+        index->farbcode = SDL_MapRGBA(pf, 0, 0, 0, 255);
       } else if (kommtString("white")) {
-        farben[index] = qRgb(255, 255, 255) | 0xff000000;
+        index->farbcode = SDL_MapRGBA(pf, 255, 255, 255, 255);
       } else if (kommtString("Background")) {
         // Farbe vom Level-Hintergrund (fr Explosion)
-        farben[index] = ld->mHintergrundFarbe.rgb() | 0xff000000;
-      } else if (kommtString("FontDark")) {
-        // Farbe der Schrift in dem Level (fr Punkt-Ziffern)
-        farben[index] = ld->mSchriftFarbe[schrift_dunkel].rgb() | 0xff000000;
-      } else if (kommtString("Font")) {
-        // Farbe der Schrift in dem Level (fr Punkt-Ziffern)
-        farben[index] = ld->mSchriftFarbe[schrift_normal].rgb() | 0xff000000;
-      } else if (kommtString("FontLight")) {
-        // Farbe der Schrift in dem Level (fr Punkt-Ziffern)
-        farben[index] = ld->mSchriftFarbe[schrift_hell].rgb() | 0xff000000;
+        index->farbcode = ld->mHintergrundFarbe.getPixel(pf);
+        index->hintergrund = true;
+//       } else if (kommtString("FontDark")) {
+//         // Farbe der Schrift in dem Level (fr Punkt-Ziffern)
+//         *index = ld->mSchriftFarbe[schrift_dunkel].getPixel(pf);
+//       } else if (kommtString("Font")) {
+//         // Farbe der Schrift in dem Level (fr Punkt-Ziffern)
+//         *index = ld->mSchriftFarbe[schrift_normal].getPixel(pf);
+//       } else if (kommtString("FontLight")) {
+//         // Farbe der Schrift in dem Level (fr Punkt-Ziffern)
+//         *index = ld->mSchriftFarbe[schrift_hell].getPixel(pf);
       } else {
         erwarte('#');
 	int f_r = getHex();
 	int f_g = getHex();
 	int f_b = getHex();
-	farben[index] = qRgb(f_r, f_g, f_b) | 0xff000000;
+	index->farbcode = SDL_MapRGBA(pf, f_r, f_g, f_b, 255);
       }
       liesBis('"');
     }
 
-    for (int y = 0; y < groesse_y; y++) {
-      leerUndKommentarWeg();
-      erwarte(',');
-      leerUndKommentarWeg();
-      erwarte('"');
-      uint * p = (uint*)im.scanLine(y);
-      for (int x = 0; x < groesse_x; x++) {
-	*p++ = farben[(int) (*gDatBei++)];
-      }
-      erwarte('"');
-    }
+    SDL_LockSurface(s); /* Damit wir Pixel direkt bearbeiten duerfen */
+
+    if (monochrom) {
+      if (farb) {
+        info farbe = farben.raus(tiefe);
+        for (int y=0; y<groesse_y; y++)
+          for (int x=0; x<groesse_x; x++)
+            SDLTools::getPixel32(s,x,y) = farbe.farbcode;
+        maske.fill(farbe.hintergrund);
+      };
+    } else
+      for (int y = 0; y < groesse_y; y++) {
+        leerUndKommentarWeg();
+	erwarte(',');
+	leerUndKommentarWeg();
+	erwarte('"');
+	for (int x = 0; x < groesse_x; x++) {
+          info farbe = farben.raus(tiefe);
+	  SDLTools::getPixel32(s, x, y) = farbe.farbcode;
+          maske.set_pixel(x,y,farbe.hintergrund);
+        };
+	erwarte('"');
+     }
+
+    SDL_UnlockSurface(s);
+
+    farben.loesch(tiefe);
 
     /* Angeblich drfen jetzt noch Extensionen von xpm kommen. Da kmmern wir
        uns nicht weiter drum. */
@@ -371,16 +443,26 @@ bool ladXPM(QImage & im, __String na) {
     if (gDatAnfang) free(gDatAnfang);
 
     if (gDebug)
-      fprintf(stderr, _("Fast xpm loading did not work for \"%s\":\n%s\n"),
-              na.data(), f.getText().data());
+      print_to_stderr(_sprintf("Fast xpm loading did not work for \"%s\":\n%s\n",
+			       na.data(), f.getText().data()));
     
-    if (gDateiGezippt) {
+    /* Ich habe im Nachfolgenden IMG_Load-Verwendung auskommentiert:
+       IMG_Load scheint nicht zu funkionieren (segmentation fault).
+       Oder vielleicht funktioniert es auch, aber es reagiert mit segfault
+       auf Dateien, die es nicht lesen kann. Deshalb: Wenn wir die Datei
+       nicht selbst lesen koennen, dann lieber den Fehler ausgeben als
+       IMG_Load aufrufen. */
+//    if (gDateiGezippt) {
       /* Bei einer gezippten Datei geht fallBackLaden nicht... */
+      // TRANSLATORS: This is to prepend an error message
       throw Fehler("File \"%s.gz\": %s\n", na.data(), f.getText().data());
-    } else
-      return fallBackLaden(im, na);
+//    } else {
+      /* Die sdl-Laderoutine verwenden. Waere eh zu pruefen ob die nicht besser ist als
+         unsere eigene */
+//      return IMG_Load(na.data());
+//    }
   }
   
   
-  return true;
+  return s;
 }
